/*
    udos: micro disk OS.
    Copyright (c) 2003, 2004   Katsuhiko Gondow (gondow@cs.titech.ac.jp)
    $Id: cpu.c,v 1.11 2004/06/23 07:59:03 gondow Exp $
*/
/* ---------------------------------------------------------------------- */
#include "cpu.h"
#include "vga.h"
#include "misc.h"
#include "paging.h"
#include "boot32.h"
#include "task.h"

/* ---------------------------------------------------------------------- */
struct GDT gdt [MAX_GDT];
struct TSS kernel_tss;
/* ---------------------------------------------------------------------- */
static struct IDT idt [MAX_GDT];
static struct selector gdt_selector, idt_selector;
/* ---------------------------------------------------------------------- */
static void set_trap_gate (int i, void (*handler)(void));
static void set_intr_gate (int i, void (*handler)(void));
static void set_syscall_gate (int i, void (*handler)(void));
static void default_exc_handler (uint32_t error);
static void default_intr_handler (void);
/* ---------------------------------------------------------------------- */
static void
set_trap_gate (int i, void (*handler)(void))
{
    idt[i].word[0] = (unsigned int)handler & 0xFFFF;
    idt[i].word[1] = 0x08;
    idt[i].word[2] = 0x8F00; /* 32bit trap gate, copy count=0 */
    idt[i].word[3] = ((unsigned int)handler >> 16) & 0xFFFF;
}

static void
set_intr_gate (int i, void (*handler)(void))
{
    set_trap_gate (i, handler);
    idt[i].word[2] = 0x8E00; /* 32bit intr gate, copy count=0 */
}


static void
set_syscall_gate (int i, void (*handler)(void))
{
    set_trap_gate (i, handler);
    idt[i].word[2] |= 0x6000; /* DPL=3, see p.157 in i386-3j.pdf */
}

static void
default_exc_handler (uint32_t error)
{
    printk ("exception occurred: error=%x\n", error);
    dump_stack ();
    asm volatile ("cli; 1: hlt; jmp 1b");
    asm volatile ("leave; iret");
}

static void
default_intr_handler (void)
{
    printk ("intterrupt occurred\n");    
    dump_stack ();
    asm volatile ("cli; 1: hlt; jmp 1b");
    asm volatile ("leave; iret");
}

/* ---------------------------------------------------------------------- */
void
setup_GDT_again (void)
{
    /* 0x00, null selector */
    gdt[0].word[0] = 0x0000; gdt[0].word[1] = 0x0000;
    gdt[0].word[2] = 0x0000; gdt[0].word[3] = 0x0000;

    /* 0x08, code/DPL=0, kernel code */
    gdt[1].word[0] = 0xFFFF; gdt[1].word[1] = 0x0000;
    gdt[1].word[2] = 0x9A00; gdt[1].word[3] = 0x00CF;

    /* 0x10, data/DPL=0, kernel data/stack */
    gdt[2].word[0] = 0xFFFF; gdt[2].word[1] = 0x0000;
    gdt[2].word[2] = 0x9200; gdt[2].word[3] = 0x00CF;

    /* 0x18, code/DPL=3, user code */
    gdt[3].word[0] = 0xFFFF; gdt[3].word[1] = 0x0000;
    gdt[3].word[2] = 0xFA00; gdt[3].word[3] = 0x00CF;

    /* 0x20, data/DPL=3, user data/stack */
    gdt[4].word[0] = 0xFFFF; gdt[4].word[1] = 0x0000;
    gdt[4].word[2] = 0xF200; gdt[4].word[3] = 0x00CF;

    {
	/* 0x28, kernel TSS */
    	uint32_t addr  = (uint32_t)&kernel_tss;
	uint32_t limit = sizeof (kernel_tss);

	gdt[5].word[0] = limit & 0xFFFF;
	gdt[5].word[1] = addr  & 0xFFFF;
	gdt[5].word[2] = 0x8900 | ((addr >> 16) & 0xFF);
	gdt[5].word[3] = ((addr  >> 16) & 0xFF00)
	               | ((0x0000)      & 0x00F0)
	               | ((limit >> 16) & 0x000F);

	/* xxx: initialize kernel TSS */
	kernel_tss.ss0  = 0x10; /* kernel data selector */
	kernel_tss.esp0 = KERNEL_STACK_ADDR;
    	kernel_tss.cr3 = (uint32_t) &kernel_page_dir; /* defined in boot32.S */
	kernel_tss.ldt = 0x00; /* null selector */
	kernel_tss.trace  = 0; /* no trap */
	kernel_tss.iobase = 0; /* empty I/O permission map */
    }

    /* gdt[6]...gdt[MAX_GDT] for TSS descriptors */

    gdt_selector.limit = sizeof (gdt) - 1;
    gdt_selector.base  = (unsigned int) gdt;

    /* we need to load again all segment registers, 
       to renew hidden part of them. See also p.71 in i386-3j.pdf. */
    asm volatile ("lgdt %0         \n"
	          "ljmp $0x8,  $1f \n"	   /* reload %cs */
	          "1:		 \n"
	          "movw $0x10, %%ax\n"
	          "movw %%ax,  %%ds\n"
	          "movw %%ax,  %%es\n"
	          "movw %%ax,  %%fs\n"
	          "movw %%ax,  %%gs\n"
	          "movw %%ax,  %%ss"  ::"m"(gdt_selector));

    /* initial TSS is loaded to TR register */
#if 1
    asm volatile ("movw %0,    %%ax\n"
		  "ltr  %%ax"::"g"(0x28));
#endif
}

void
setup_IDT_again (void)
{
    int i;

    /* setup default exception handler */
    for (i = 0; i < 0x20; i++) {
	set_trap_gate (i, (void (*)(void))default_exc_handler);
    }

    /* setup default interruption handler */
    for (i = 0x20; i < 0x30; i++) {
	set_intr_gate (i, default_intr_handler);
    }

    /* setup event-specific handlers */
    {
	set_intr_gate (0x20, pit_intr_handler_entry);
	set_intr_gate (0x21, kbd_intr_handler_entry);
	set_intr_gate (0x26, fdc_intr_handler_entry);
	set_intr_gate (0x0E, page_fault_handler_entry);
	set_intr_gate (0x0A, tss_fault_handler_entry); /* for debug */
	set_intr_gate (0x0D, gp_fault_handler_entry);  /* for debug */
	set_syscall_gate (0x80, system_call_entry);
	set_syscall_gate (0x81, test_entry);
    }

    idt_selector.limit = sizeof (idt) - 1;
    idt_selector.base  = (unsigned int) idt;
    asm volatile ("lidt %0"::"m"(idt_selector));
}
/* ---------------------------------------------------------------------- */
