/*
    udos: micro disk OS.
    Copyright (c) 2003, 2004   Katsuhiko Gondow (gondow@cs.titech.ac.jp)
    $Id: fat12.c,v 1.34 2005/07/31 06:32:32 gondow Exp gondow $
*/

/*
    Limitations.
    - No handling long file names.
    - No defragmentation, no garbage collection.

    Note: there are two ways to access directory entries (`dirent' for short),
    	 (1) FAT12_read/FAT12_write for sequential access, and
	 (2) read_dirent/write_dirent for random access.
    
    Note: 10 reasons why FAT12 is not beautiful.
    - root directory is not a chain of clusters, but a contiguous region.
    - all size entries in dirent for directories contain the value 0,
      so we need to count their cluster chains to obtain their real sizes.
    - The values of ROOT_DIR_CLUS_NUM and CLUSTER_FREE are the same,
      which is very confusing.
    - delayed freeing of unused clusters.
    - the mixture of short and long names.
    - ... (omitted)
*/
#if defined(KERNEL)
#include "misc.h"   /* for panic, memcpy */
#include "fd.h"     /* for FDC_init, FDC_read, FDC_write */
#include "vga.h"    /* for kbd_get_char, kbd_get_line */
#include "paging.h" /* for kmalloc, kfree */
#include "task.h"   /* for process_table */
#define  malloc kmalloc
#define  free   kfree
#define  printf printk
#define  perror panic
#else  /* KERNEL */
#include <stdio.h>
#include <stdint.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>
#endif /* KERNEL */

#include "fat12.h"

#define min(data1, data2)   (((data1) > (data2)) ? (data2) : (data1))
/* ---------------------------------------------------------------------- */
enum {
    SECTOR_SIZE    = 512,
    DIRENT_SIZE    = 32,
    CLUSTER_FREE   = 0x0000,
    CLUSTER_BAD    = 0x0FF7,
    FIRST_CLUSTER  = 2,
    DIRENT_UNUSED  = 0xE5,
    DIRENT_END_DIR = 0x00,
};

/* BPB (BIOS parameter block) in PBR (partition boot record) */
struct FAT_BPB {
    int  BytsPerSec;
    int  SecPerClus;
    int  RsvdSecCnt;
    int  NumFATs;
    int  RootEntCnt;
    int  TotSec16;
    int  FATSz16;
    int  SecPerTrk;
    int  NumHeads;
    int  HiddSec;
};

struct FAT_info {
    int  RootDirSectors;
    int  FirstDataSector;
    int  DataSec;
    int  CountofClusters; /* excluding the two reserved clusters */
    int  FirstRootDirSecNum;
    int  BytsPerClus;
    int  EndOfCluster;
    int  FATStart, FATSize;
    int  RootStart, RootSize;
    struct FAT_dirent DummyRootDirent;
    uint8_t *FAT;
    uint8_t *RootDir;
};
/* ---------------------------------------------------------------------- */
#if !defined(KERNEL)
static int    FAT_current_dir = ROOT_DIR_CLUS_NUM;
static FILE   *image_fp, *file_fp;
static int    is_dump = 0;
#endif /* !KERNEL */

static struct FAT_BPB  FAT_BPB;
static struct FAT_info FAT_info;
/* ---------------------------------------------------------------------- */
static void  *emalloc (int size);

static void   fs_read  (void *buf, int size, int offset);
static void   fs_write (void *buf, int size, int offset);

static void   read_FAT12_PBR (void);
static void   set_FAT12_info (void);
static void   read_FAT12  (void);
static void   write_FAT12 (void);
static void   read_FAT12_root_directory (void);
static void   write_FAT12_root_directory (void);
static void   read_cluster  (void *buf, int clus_num);
static void   write_cluster (void *buf, int clus_num);

static void   decode_dirent (uint8_t *ptr, struct FAT_dirent *dirent);
static void   encode_dirent (uint8_t *ptr, struct FAT_dirent *dirent);
static void   read_dirent   (int clus_num, int clus_offset,
			     struct FAT_dirent *dirent);
static void   write_dirent  (int clus_num, int clus_offset,
			     struct FAT_dirent *dirent);
static int    search_dirent (int dir_clus_num, const char *filename,
			     struct FAT_dirent *dirent);
static int    create_dirent (const char *path, struct FAT_dirent *new_dirent,
			     int attr, int clus_num);
static int    path_to_dirent (const char *path, struct FAT_dirent *dirent);

static int    get_FAT12_entry (int clus_num);
static void   set_FAT12_entry (int clus_num, int value);
static int    alloc_cluster (int start);
static void   alloc_cluster_chain (struct FAT_dirent *dirent, int last_offset);
static void   free_cluster_chain (int clus_num);

static int    FAT12_read_write_root_dirent (struct file *fp, void *buf,
					    int is_read);
static int    FAT12_read_write (struct file *fp, void *buf, int count,
				int is_read);

static int    is_root_directory    (struct FAT_dirent *dirent);
static int    is_end_of_directory  (struct FAT_dirent *dirent);
static int    is_invalid_dirent    (struct FAT_dirent *dirent);
static void   normalize_name       (const char *from, char *to);
static int    get_filesize         (struct FAT_dirent *dirent);
static void   calc_clus_num_offset (struct file *fp, int *clus_num,
				    int *clus_offset);
static char  *divide_path_right     (const char *path, char *file);
static const char *divide_path_left (const char *path, char *file);
#if defined(KERNEL)
static int    toupper (int c);
#else /* !defined(KERNEL) */
static int    efseek (FILE *fp, long offset, int whence);
static size_t efread (void *buf, size_t size, size_t count, FILE *fp);
static size_t efwrite (const void *buf, size_t size, size_t count, FILE *fp);
static void   FAT12_dump_fileinfo (struct FAT_dirent *dirent, int depth);
static void   FAT12_dump_directory (struct FAT_dirent *dot_dirent, int depth);
static void   FAT12_dump (void);
static void   FAT12_add  (char *filename);
#endif /* KERNEL */
/* ---------------------------------------------------------------------- */
static void *
emalloc (int size)
{ 
    void *p = malloc (size);
    if (p == NULL) perror ("malloc");
    return p;
}

static void
fs_read  (void *buf, int size, int offset)
{
#if defined(KERNEL)
    assert (size   % SECTOR_SIZE == 0);
    assert (offset % SECTOR_SIZE == 0);
    buffer_readn (offset / SECTOR_SIZE, size / SECTOR_SIZE, buf);
#else
    efseek (image_fp, offset, SEEK_SET);
    efread (buf, size, 1, image_fp);
#endif /* !KERNEL */
}

static void
fs_write (void *buf, int size, int offset)
{
#if defined(KERNEL)
    assert (size   % SECTOR_SIZE == 0);
    assert (offset % SECTOR_SIZE == 0);
    buffer_writen (offset / SECTOR_SIZE, size / SECTOR_SIZE, buf);
#else
    efseek  (image_fp, offset, SEEK_SET);
    efwrite (buf, size, 1, image_fp);
#endif
}

static void
read_FAT12_PBR (void)
{
    uint8_t buf [SECTOR_SIZE];

    fs_read (buf, SECTOR_SIZE, 0);
    /* note: *(uint16_t *)&buf [11] may cause alignment problems.
       FAT file system is all "little endian", so this solution is correct. */
    FAT_BPB.BytsPerSec = buf [11] + (buf [12] << 8);
    FAT_BPB.SecPerClus = buf [13];
    FAT_BPB.RsvdSecCnt = buf [14] + (buf [15] << 8);
    FAT_BPB.NumFATs    = buf [16];
    FAT_BPB.RootEntCnt = buf [17] + (buf [18] << 8);
    FAT_BPB.TotSec16   = buf [19] + (buf [20] << 8);
    FAT_BPB.FATSz16    = buf [22] + (buf [23] << 8);
    FAT_BPB.SecPerTrk  = buf [24] + (buf [25] << 8);
    FAT_BPB.NumHeads   = buf [26] + (buf [27] << 8);
    FAT_BPB.HiddSec    = (buf [28] + (buf [29] << 8)
			  + (buf [30] << 16) + (buf [31] << 24));
    assert (FAT_BPB.FATSz16 != 0);
}

static void
set_FAT12_info (void)
{
    FAT_info.RootDirSectors =
	((FAT_BPB.RootEntCnt * DIRENT_SIZE) + (FAT_BPB.BytsPerSec - 1))
	/ FAT_BPB.BytsPerSec;
    FAT_info.FirstDataSector =
	FAT_BPB.RsvdSecCnt + (FAT_BPB.NumFATs * FAT_BPB.FATSz16)
	+ FAT_info.RootDirSectors;
    FAT_info.DataSec = FAT_BPB.TotSec16 -
	(FAT_BPB.RsvdSecCnt + (FAT_BPB.NumFATs * FAT_BPB.FATSz16)
	 + FAT_info.RootDirSectors);
    FAT_info.CountofClusters = FAT_info.DataSec / FAT_BPB.SecPerClus;
    FAT_info.FirstRootDirSecNum = 
	FAT_BPB.RsvdSecCnt + (FAT_BPB.NumFATs * FAT_BPB.FATSz16); 
    FAT_info.BytsPerClus = FAT_BPB.BytsPerSec * FAT_BPB.SecPerClus;

    FAT_info.FATStart = FAT_BPB.RsvdSecCnt * FAT_BPB.BytsPerSec;
    FAT_info.FATSize  = FAT_BPB.FATSz16 * FAT_BPB.BytsPerSec;
    FAT_info.RootStart = FAT_info.FirstRootDirSecNum * FAT_BPB.BytsPerSec;
    FAT_info.RootSize  = FAT_info.RootDirSectors * FAT_BPB.BytsPerSec;

    /* FAT12 satisfies this condition */
    assert (0 < FAT_info.CountofClusters && FAT_info.CountofClusters < 4085);

    /* FAT_info.DummyRootDirent init. */
    memcpy (FAT_info.DummyRootDirent.name, "(ROOT)     ", FILENAME_LEN + 1);
    FAT_info.DummyRootDirent.attr     = ATTR_DIRECTORY;
    FAT_info.DummyRootDirent.clus_num = ROOT_DIR_CLUS_NUM;
    FAT_info.DummyRootDirent.size     = 0;

    FAT_info.FAT     = emalloc (FAT_info.FATSize);
    FAT_info.RootDir = emalloc (FAT_info.RootSize);
}

static void
read_FAT12 (void)
{
    fs_read (FAT_info.FAT, FAT_info.FATSize, FAT_info.FATStart);
    FAT_info.EndOfCluster = get_FAT12_entry (1);
}

static void
write_FAT12 (void)
{
    int i, size = FAT_info.FATSize;
    for (i = 0; i < FAT_BPB.NumFATs; i++)
	fs_write (FAT_info.FAT, size, FAT_info.FATStart + size * i);
}

static void
read_FAT12_root_directory (void)
{
    fs_read (FAT_info.RootDir, FAT_info.RootSize, FAT_info.RootStart);
}

static void
write_FAT12_root_directory (void)
{
    fs_write (FAT_info.RootDir, FAT_info.RootSize, FAT_info.RootStart);
}

static void
read_cluster (void *buf, int clus_num)
{
    int offset = (FAT_info.FirstDataSector +
		  (clus_num - 2) * FAT_BPB.SecPerClus) * FAT_BPB.BytsPerSec;
    fs_read  (buf, FAT_info.BytsPerClus, offset);
}

static void
write_cluster (void *buf, int clus_num)
{
    int offset = (FAT_info.FirstDataSector +
		  (clus_num - 2) * FAT_BPB.SecPerClus) * FAT_BPB.BytsPerSec;
    fs_write  (buf, FAT_info.BytsPerClus, offset);
}

static void
decode_dirent (uint8_t *ptr, struct FAT_dirent *dirent)
{
    memcpy (dirent->name, ptr, FILENAME_LEN);
    dirent->name [FILENAME_LEN] = '\0';
    dirent->attr = ptr [11];
    dirent->size = ptr [28] | (ptr [29] << 8)
	    | (ptr [30] << 16) | (ptr [31] << 24);
    dirent->clus_num = ptr [26] | (ptr [27] << 8);
    /* Note: fatgen103.pdf says ptr[20] and ptr[21] are always 0
       for FAT12/FAT16, but not in some implementation. So, we
       ignore ptr[20] and ptr[21]. | (ptr [20] << 16) | (ptr [21] << 24); */
}

static void
encode_dirent (uint8_t *ptr, struct FAT_dirent *dirent)
{
    memcpy (ptr, dirent->name, FILENAME_LEN);
    ptr [11] = dirent->attr;
    ptr [28] = dirent->size             & 0xFF;
    ptr [29] = (dirent->size >> 8)      & 0xFF;
    ptr [30] = (dirent->size >> 16)     & 0xFF;
    ptr [31] = (dirent->size >> 24)     & 0xFF;
    ptr [26] = dirent->clus_num         & 0xFF;
    ptr [27] = (dirent->clus_num >> 8)  & 0xFF;
    ptr [20] = (dirent->clus_num >> 16) & 0xFF;
    ptr [21] = (dirent->clus_num >> 24) & 0xFF;
    ptr [12] = 0; /* initialize reserved area */

    /* initialized time stamps with the epoch time */
    ptr [13] = 0;                /* CrtTimeTenth */
    ptr [14] = ptr [15] = 0;     /* CrtTime */
    ptr [16] = ptr [17] = 0;     /* CrtDate */
    ptr [18] = ptr [19] = 0;     /* LstAccDate */
    ptr [22] = ptr [23] = 0;     /* WrtTime */
    ptr [24] = ptr [25] = 0;     /* WrtDate */
}

static void
read_dirent (int clus_num, int clus_offset, struct FAT_dirent *dirent)
{
    assert (clus_offset % DIRENT_SIZE == 0);
    if (clus_num == ROOT_DIR_CLUS_NUM) {
	decode_dirent (FAT_info.RootDir + clus_offset, dirent);
    } else {
	uint8_t *buf = emalloc (FAT_info.BytsPerClus);
	read_cluster (buf, clus_num);
	decode_dirent (buf + clus_offset, dirent);
	free (buf);
    }
    dirent->entry.clus_num    = clus_num;
    dirent->entry.clus_offset = clus_offset;
}

static void
write_dirent (int clus_num, int clus_offset, struct FAT_dirent *dirent)
{
    assert (clus_offset % DIRENT_SIZE == 0);
    if (clus_num == ROOT_DIR_CLUS_NUM) {
	encode_dirent (FAT_info.RootDir + clus_offset, dirent);
    } else {
	uint8_t *buf = emalloc (FAT_info.BytsPerClus);
	read_cluster (buf, clus_num);
	encode_dirent (buf + clus_offset, dirent);
	write_cluster (buf, clus_num);
	free (buf);
    }
}

static int
search_dirent (int dir_clus_num, const char *filename,
	       struct FAT_dirent *dirent)
{
    struct file file;

    if ((dir_clus_num == ROOT_DIR_CLUS_NUM)
	&& (!strncmp (filename, DIR_DOT, FILENAME_LEN)
	    || !strncmp (filename, DIR_DOTDOT, FILENAME_LEN))) {
	*dirent = FAT_info.DummyRootDirent;
	return 0;
    }

    /* to call FAT12_readdir without FAT12_open */
    file.flag                     = O_RDONLY;
    file.current_offset           = 0;
    file.current_clus_num         = dir_clus_num;
    file.current_clus_offset      = 0;
    memcpy (file.dirent.name, filename, FILENAME_LEN + 1);
    file.dirent.attr              = ATTR_DIRECTORY;
    file.dirent.clus_num          = dir_clus_num;
    file.dirent.size              = 0;
    file.dirent.entry.clus_num    = -1;
    file.dirent.entry.clus_offset = -1;

    while (FAT12_readdir (&file, dirent, 1) > 0) {
	if (!strncmp (filename, dirent->name, FILENAME_LEN))
	    return 0;
    }
    return -1;
}

static int
create_dirent (const char *path, struct FAT_dirent *new_dirent,
	       int attr, int clus_num)
{
    char *dir, filename [FILENAME_LEN + 1];
    struct file file;
    int ret;

    new_dirent->size     = 0;
    new_dirent->attr     = attr;
    new_dirent->clus_num = clus_num;

    ret = FAT12_open (&file, path, O_RDONLY);
    if (ret != -1) {
	printf ("create_dirent: already exists: %s\n", path);
	goto error;
    }
    FAT12_close (&file);

    dir = divide_path_right (path, filename);
    ret = FAT12_open (&file, dir, O_RDWR);
    if (ret == -1) {
	printf ("create_dirent: No such directory: %s\n", dir);
	goto error;
    }
    if (!is_directory (&file.dirent)) {
	printf ("create_dirent: Not directory: %s\n", dir);
	goto error;
    }

    memcpy (new_dirent->name, filename, FILENAME_LEN + 1);
    ret = FAT12_writedir (&file, new_dirent, 1);
    FAT12_close (&file);
    free (dir);
    return 0;
 error:
    FAT12_close (&file);
    free (dir);
    return -1;
}

/* like namei in BSD */
static int
path_to_dirent (const char *path, struct FAT_dirent *dirent)
{
    const char *dir = path;
    int ret, clus_num;

#if defined(KERNEL)
    clus_num = (*path== '/' ? ROOT_DIR_CLUS_NUM 
		: process_table [sys_getpid ()].current_dir);
#else
    extern int FAT_current_dir;
    clus_num = (*path== '/' ? ROOT_DIR_CLUS_NUM : FAT_current_dir);
#endif

    if (!strcmp (path, "/")) {
	*dirent = FAT_info.DummyRootDirent;
	return 0;
    }
    while (1) {
	char file [FILENAME_LEN + 1];
	dir = divide_path_left (dir, file);
	ret = search_dirent (clus_num, file, dirent);
	if (ret == -1)
	    return -1;
	clus_num = dirent->clus_num;
	if (dir == NULL)
	    break;
    }
    return 0;
}

static int
get_FAT12_entry (int clus_num)
{
    int FAT_entry;
    int FATOffset = clus_num + (clus_num / 2);

    if (clus_num % 2 == 0) { /* cluster number is even */
	FAT_entry = (FAT_info.FAT [FATOffset] & 0xFF)
		    | ((FAT_info.FAT [FATOffset + 1] << 8) & 0xF00);
    } else {                 /* cluster number is odd  */
	FAT_entry = ((FAT_info.FAT [FATOffset] >> 4) & 0xF)
		    | ((FAT_info.FAT [FATOffset + 1] << 4) & 0xFF0);
    }
    return FAT_entry;
}

static void
set_FAT12_entry (int clus_num, int value)
{
    int FATOffset = clus_num + (clus_num / 2);

    if (clus_num % 2 == 0) { /* cluster number is even */
	FAT_info.FAT [FATOffset] = value & 0xFF;
	FAT_info.FAT [FATOffset + 1] = (FAT_info.FAT [FATOffset + 1] & 0xF0)
	                               | ((value >> 8) & 0x0F);
    } else {                 /* cluster number is odd  */
	FAT_info.FAT [FATOffset] = (FAT_info.FAT [FATOffset] & 0x0F)
	                           | ((value << 4) & 0xF0);
	FAT_info.FAT [FATOffset + 1] = (value >> 4) & 0xFF;
    }
}

/* return value: clus_num (>0) = first free cluster */
static int
alloc_cluster (int start)
{
    int i;
    assert (start >= FIRST_CLUSTER);
    for (i = start; i < FAT_info.CountofClusters + FIRST_CLUSTER; i++) {
	if (get_FAT12_entry (i) == CLUSTER_FREE) {
	    set_FAT12_entry (i, FAT_info.EndOfCluster);
	    return i;
	}
    }
    assert (0 && "no free cluster left");
    return 0; /* dummy */
}

static void
alloc_cluster_chain (struct FAT_dirent *dirent, int last_offset)
{
    int clus_size = FAT_info.BytsPerClus;
    int filesize  = get_filesize (dirent);
    int clus_num;
    int curr_clus = (filesize    + clus_size - 1) / clus_size;
    int new_clus  = (last_offset + clus_size - 1) / clus_size;
    int n         = new_clus - curr_clus;

    assert (n > 0);
    /* search last cluster */
    if (dirent->clus_num == CLUSTER_FREE) {
	clus_num = dirent->clus_num = alloc_cluster (FIRST_CLUSTER);
	n--;
    } else {
	for (clus_num = dirent->clus_num;
	     get_FAT12_entry (clus_num) != FAT_info.EndOfCluster;
	     clus_num = get_FAT12_entry (clus_num))
	    ; /* empty body */
    }

    while (n--) {
	int next = alloc_cluster (clus_num + 1);
	set_FAT12_entry (clus_num, next);
	clus_num = next;
    }

    if (!is_directory (dirent)) 
	dirent->size = last_offset;
    write_dirent (dirent->entry.clus_num, dirent->entry.clus_offset, dirent);
}

/* Note: for now, do not use this function for unused dirents.
   Unused directory entries (0xE5) can contain non-zero cluster numbers,
   which can help disk utilities recover the files. But those non-zero
   cluster numbers can be shared with other files, so simply freeing
   the clusters results in file system crash. For simplicity, udos does
   not try to collect such garbages although it is a waste of space. */
static void
free_cluster_chain (int clus_num)
{
    while (1) {
	int next;
	if (clus_num == CLUSTER_FREE)          break;
	if (clus_num == FAT_info.EndOfCluster) break;
	next = get_FAT12_entry (clus_num);
	set_FAT12_entry (clus_num, CLUSTER_FREE);
	clus_num = next;
    }
}

static int
FAT12_read_write_root_dirent (struct file *fp, void *buf, int is_read)
{
    /* FAT12 root has the fixed size */
    if (fp->current_offset + DIRENT_SIZE > get_filesize (&fp->dirent)) {
	return 0;
    }

    if (is_read)
	memcpy (buf, FAT_info.RootDir + fp->current_offset, DIRENT_SIZE);
    else
	memcpy (FAT_info.RootDir + fp->current_offset, buf, DIRENT_SIZE);
    fp->current_offset += DIRENT_SIZE;
    calc_clus_num_offset (fp, &fp->current_clus_num, &fp->current_clus_offset);
    return DIRENT_SIZE;
}

static int
FAT12_read_write   (struct file *fp, void *buf, int count, int is_read)
{
    uint8_t *SecBuff;
    uint8_t *p = buf;
    int clus_offset, clus_num;
    int n_left, size, filesize;

#if 0 /* xxx */
    printf ("current_offset=%d, current_clus_num=%d, current_clus_offset=%d, is_root_directory=%d\n", 
	    fp->current_offset, fp->current_clus_num, fp->current_clus_offset,
	    is_root_directory (&fp->dirent));
#endif
    filesize = get_filesize (&fp->dirent);
    /* decrease 'count' to avoid over-reading */
    if (is_read && fp->current_offset + count > filesize) {
	count = filesize - fp->current_offset;
    }

    /* allocate clusters to allow over-writing */
    if (!is_read && fp->current_offset + count > filesize) {
	alloc_cluster_chain (&fp->dirent, fp->current_offset + count);
    }

    if (is_root_directory (&fp->dirent))
	return FAT12_read_write_root_dirent (fp, buf, is_read);

    SecBuff = emalloc (FAT_info.BytsPerClus);
    calc_clus_num_offset (fp, &clus_num, &clus_offset);
    n_left = count;
    size = min (n_left, FAT_info.BytsPerClus - clus_offset);

    /* transfer data cluster-by-cluster or less */
    while (n_left > 0) {
	if (is_read) {
	    read_cluster (SecBuff, clus_num);
	    memcpy (p, SecBuff + clus_offset, size);
	} else {
	    if (clus_offset != 0 || size != FAT_info.BytsPerClus)
		read_cluster (SecBuff, clus_num);
	    memcpy (SecBuff + clus_offset, p, size);
	    write_cluster (SecBuff, clus_num);
	}

	p      += size;
	n_left -= size;
	clus_offset = 0;
	clus_num = get_FAT12_entry (clus_num);
	size = min (n_left, FAT_info.BytsPerClus);
    }

    fp->current_offset     += count;
    calc_clus_num_offset (fp, &fp->current_clus_num, &fp->current_clus_offset);

    free (SecBuff);
    return count;
}

static int
is_root_directory (struct FAT_dirent *dirent)
{
    return is_directory (dirent) && (dirent->clus_num == ROOT_DIR_CLUS_NUM);
}

static int
is_end_of_directory (struct FAT_dirent *dirent)
{
    return (uint8_t)dirent->name [0] == DIRENT_END_DIR;
}

static int
is_invalid_dirent (struct FAT_dirent *dirent)
{
    if ((uint8_t)dirent->name [0] == DIRENT_UNUSED)
	return 1;      /* unused entry */
    if ((dirent->attr & ATTR_LONG_NAME_MASK) == ATTR_LONG_NAME)
	return 1;      /* entry for long file name */
    if ((dirent->attr & (ATTR_VOLUME_ID | ATTR_DIRECTORY)) == ATTR_VOLUME_ID)
	return 1;      /* entry for volume name */
    return 0;
}

/* conversion like "udos.bin" -> "UDOS    BIN" */
static void
normalize_name (const char *from, char *to)
{
    int i;

    memset (to, ' ', FILENAME_LEN);
    to [FILENAME_LEN] = '\0';

    if (!strncmp (from, "..", 2)) {
	memcpy (to, DIR_DOTDOT, FILENAME_LEN + 1);
	return;
    }
    if (!strncmp (from, ".", 1)) {
	memcpy (to, DIR_DOT,    FILENAME_LEN + 1);
	return;
    }

    for (i = 0; i < 8; i++) {
	if (*from == '.')  break;
	if (*from == '\0') return;
	to [i] = toupper (*from++);
    }
    while (*from != '\0' && *from++ != '.')
	;
    for (i = 8; i < FILENAME_LEN; i++) {
	if (*from == '\0') break;
	to [i] = toupper (*from++);
    }
}

static int
get_filesize (struct FAT_dirent *dirent)
{
    int i, n_clus;
    if (!is_directory (dirent))
	return dirent->size;
    if (is_root_directory(dirent))
	return FAT_info.RootDirSectors * FAT_BPB.BytsPerSec;
    if (dirent->clus_num == CLUSTER_FREE) /* directories but root */
	return 0;
    /* counting clusters used for the directory */
    n_clus = 0;
    for (i = dirent->clus_num; i != FAT_info.EndOfCluster;
	 i = get_FAT12_entry (i)) {
	n_clus++;
    }
    return n_clus * FAT_info.BytsPerClus;
}

static void
calc_clus_num_offset (struct file *fp, int *clus_num, int *clus_offset)
{
    int clus_nth;

    if (is_root_directory (&fp->dirent)) {
	*clus_offset = fp->current_offset;
	return;
    }
    clus_nth = fp->current_offset / FAT_info.BytsPerClus;
    *clus_offset = fp->current_offset % FAT_info.BytsPerClus;
    *clus_num    = fp->dirent.clus_num;
    while (clus_nth--)
	*clus_num = get_FAT12_entry (*clus_num);
}

/* "/usr/bin/ls" -> return="/usr/bin", file="LS         " */
static char *
divide_path_right (const char *path, char *file)
{
    char *s;
    char *path2 = emalloc (strlen (path) + 1);

    memcpy (path2, path, strlen (path) + 1);
    s = strrchr (path2, '/');
    if (s == NULL) {
	normalize_name (path2, file);
	free (path2);
	return strdup (DIR_DOT);
    } else {
	*s = '\0';
	normalize_name (s + 1, file);
	if (*path2 == '\0') {
	    free (path2);
	    return strdup (DIR_ROOT);
	} else {
	    return path2;
	}
    }
}

/* "/usr/bin/ls" or "usr/bin/ls" -> file="USR        ", return="bin/ls".
   Note: divide_path_{right, left} are not symmetric. */
static const char *
divide_path_left (const char *path, char *file)
{
    char *s;
    char short_filename [FILENAME_LEN + 1];

    if (*path == '/') path++;
    s = strchr (path,  '/');
    if (s == NULL) {
	normalize_name (path, file);
	return NULL;
    } else {
	memset (short_filename, ' ', FILENAME_LEN);
	memcpy (short_filename, path, s - path);
        short_filename [s - path] = '\0'; 
	normalize_name (short_filename, file);
	return s + 1;
    }
}

#if defined(KERNEL)
static int
toupper (int c)
{
    if ('a' <= c && c <= 'z')
	return c + 'A' - 'a';
    else
	return c;
}
#else /* !defined(KERNEL) */
static int
efseek (FILE *fp, long offset, int whence)
{
    int ret = fseek (fp, offset, whence);
    if (ret == EOF) perror ("efseek");
    return ret;
}

static size_t
efread (void *buf, size_t size, size_t count, FILE *fp)
{
    int ret = fread (buf, size, count, fp);
    if (ferror (fp) != 0) perror ("efread");
    return ret;
}

static size_t
efwrite (const void *buf, size_t size, size_t count, FILE *fp)
{
    int ret = fwrite (buf, size, count, fp);
    if (ferror (fp) != 0) perror ("efwrite");
    return ret;
}

static void
FAT12_dump_fileinfo (struct FAT_dirent *dirent, int depth)
{
    int clus_num;

    printf ("%*s%s, attr=0x%x, size=%d, clusters=%x, ", depth*2, "",
	    dirent->name, dirent->attr, dirent->size, dirent->clus_num);
    clus_num = dirent->clus_num;
    while (clus_num != FAT_info.EndOfCluster && clus_num != 0) {
	clus_num = get_FAT12_entry (clus_num);
	printf ("%x, ", clus_num);
    }
    putchar ('\n');
}

static void
FAT12_dump_directory (struct FAT_dirent *dot_dirent, int depth)
{
    int i, j;
    struct FAT_dirent dirent;

    for (i = dot_dirent->clus_num; i != FAT_info.EndOfCluster;
	 i = get_FAT12_entry (i)) {
	/* for each dirent */
	for (j = 0; j < FAT_info.BytsPerClus / DIRENT_SIZE; j++) {
	    read_dirent (i, j * DIRENT_SIZE, &dirent);
#if 0
	    if (is_invalid_dirent (&dirent))                      continue;
#endif
	    if (is_end_of_directory (&dirent))                    break;
	    if (!strncmp (DIR_DOT, dirent.name, FILENAME_LEN))    continue;
	    if (!strncmp (DIR_DOTDOT, dirent.name, FILENAME_LEN)) continue;

	    FAT12_dump_fileinfo (&dirent, depth);
	    if (is_directory (&dirent))
		FAT12_dump_directory (&dirent, depth + 1);
	}
    }
}

static void
FAT12_dump (void)
{
    int i;

    /* show parameters */
    printf ("BytsPerSec = %d\n", FAT_BPB.BytsPerSec);
    printf ("SecPerClus = %d\n", FAT_BPB.SecPerClus);
    printf ("RsvdSecCnt = %d\n", FAT_BPB.RsvdSecCnt);
    printf ("NumFATs    = %d\n", FAT_BPB.NumFATs);
    printf ("RootEntCnt = %d\n", FAT_BPB.RootEntCnt);
    printf ("TotSec16   = %d\n", FAT_BPB.TotSec16);
    printf ("FATSz16    = %d\n", FAT_BPB.FATSz16);
    printf ("SecPerTrk  = %d\n", FAT_BPB.SecPerTrk);
    printf ("NumHeads   = %d\n", FAT_BPB.NumHeads);
    printf ("HiddSec    = %d\n", FAT_BPB.HiddSec);

    printf ("RootDirSectors  = %d\n", FAT_info.RootDirSectors);
    printf ("FirstDataSector = %d\n", FAT_info.FirstDataSector);
    printf ("DataSec         = %d\n", FAT_info.DataSec);
    printf ("CountofClusters = %d\n", FAT_info.CountofClusters);
    printf ("FirstRootDirSecNum = %d\n", FAT_info.FirstRootDirSecNum);
    printf ("BytsPerClus = %d\n",     FAT_info.BytsPerClus);
    printf ("EndOFCluster = %x\n",    FAT_info.EndOfCluster);

    /* show FAT table */
    printf ("      0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF"
	    "0123456789ABCDEF");
    for (i = 0; i < FAT_info.CountofClusters + 2; i++) {
	int clus_num;
	if (i % 64 == 0) printf ("\n%4x: ", i);
	clus_num = get_FAT12_entry (i);
	switch (clus_num) {
	case CLUSTER_BAD:  putchar ('!'); break;
	case CLUSTER_FREE: putchar ('.'); break;
	default:
	    if (i < 2)     putchar ('R');
	    else if (clus_num == FAT_info.EndOfCluster)
		           putchar ('E');
	    else           putchar ('*');
	    break;
	}
    }
    putchar ('\n');

    /* show directory entries in root and their cluster numbers */
    for (i = 0; i < FAT_BPB.RootEntCnt; i++) {
	struct FAT_dirent dirent;
	read_dirent (ROOT_DIR_CLUS_NUM, i * DIRENT_SIZE, &dirent);
	if (is_invalid_dirent (&dirent))   continue;
	if (is_end_of_directory (&dirent)) break;
	FAT12_dump_fileinfo (&dirent, 0);
	if (is_directory (&dirent))
	    FAT12_dump_directory (&dirent, 1);
    }
}

static void
FAT12_add  (char *filename)
{
    int nbytes, ret;
    char buf [BUFSIZ];
    struct file file;

    ret = FAT12_open (&file, filename, O_WRONLY | O_CREAT | O_TRUNC);
    assert (ret != -1);

    if ((file_fp = fopen (filename, "rb")) == NULL) {
	fprintf (stderr, "No such file: %s\n", filename);
	exit (1);
    }

    while (1) {
	nbytes = efread  (buf, 1, sizeof (buf), file_fp);
	if (nbytes <= 0) break;
	FAT12_write (&file, buf, nbytes);
    }

    FAT12_close (&file);
    fclose (file_fp);
}
#endif /* !KERNEL */
/* ---------------------------------------------------------------------- */
int
is_directory (struct FAT_dirent *dirent)
{
    return (dirent->attr & (ATTR_VOLUME_ID|ATTR_DIRECTORY)) == ATTR_DIRECTORY;
}

void
FAT12_init (void)
{
    read_FAT12_PBR ();
    set_FAT12_info ();
    read_FAT12 ();
    read_FAT12_root_directory ();
}

/* write primary FAT back to the image file. (don't forget backup FATs).
   write all buffers back to floppy disk when kernel mode. */
void
FAT12_sync (void)
{
    write_FAT12 ();
    write_FAT12_root_directory ();
#if defined(KERNEL)
    buffer_sync ();
#endif
}

int
FAT12_open   (struct file *fp, const char *path, int flag)
{
    int ret;

    ret = path_to_dirent (path, &fp->dirent);
    if (ret == -1) {
	if (flag & O_CREAT) {
	    create_dirent (path, &fp->dirent, ATTR_ARCHIVE, CLUSTER_FREE);
	} else {
	    return -1;
	}
    }
    fp->flag = flag;
    if (flag & O_APPEND) {
	fp->current_offset = get_filesize (&fp->dirent);
	fp->current_clus_num    = FAT_info.EndOfCluster;
	fp->current_clus_offset = 0;
    } else {
	fp->current_offset      = 0;
	fp->current_clus_num    = fp->dirent.clus_num;
	fp->current_clus_offset = 0;
    }
    if (flag & O_TRUNC) {
	free_cluster_chain (fp->dirent.clus_num);
	fp->dirent.clus_num     = CLUSTER_FREE;
	fp->current_clus_num    = CLUSTER_FREE;
	fp->current_clus_offset = 0;
	fp->dirent.size = 0;
	write_dirent (fp->dirent.entry.clus_num, fp->dirent.entry.clus_offset,
		      &fp->dirent);
    }
    return 0;
}

int
FAT12_close  (struct file *fp)
{
    /* do nothing for now */
    return 0;
}

int
FAT12_read   (struct file *fp, void *buf, int count)
{
    return FAT12_read_write (fp, buf, count, 1);
}

int
FAT12_write  (struct file *fp, void *buf, int count)
{
    return FAT12_read_write (fp, buf, count, 0);
}

int
FAT12_readdir (struct file *fp, struct FAT_dirent *dirent, int skip_invalid)
{
    int n_bytes;
    uint8_t ptr [DIRENT_SIZE];

    while (1) {
	dirent->entry.clus_num    = fp->current_clus_num;
	dirent->entry.clus_offset = fp->current_clus_offset;
	n_bytes = FAT12_read (fp, ptr, DIRENT_SIZE);
	decode_dirent (ptr, dirent);
	if (n_bytes < DIRENT_SIZE || is_end_of_directory (dirent))
	    break;
	if (!skip_invalid || !is_invalid_dirent (dirent))
	    return n_bytes;
    }
    return 0;
}

int
FAT12_writedir (struct file *fp, struct FAT_dirent *dirent, int skip_used)
{
    uint8_t ptr [DIRENT_SIZE];
    struct FAT_dirent dirent_tmp;
    int n_bytes, ret;

    while (1) {
	n_bytes = FAT12_read (fp, ptr, DIRENT_SIZE);
	decode_dirent (ptr, &dirent_tmp);
	if (n_bytes < DIRENT_SIZE)
	    perror ("FAT12_writedir: read error");
	if (is_end_of_directory (&dirent_tmp))
	    break;
	if (!skip_used || (uint8_t)dirent_tmp.name [0] == DIRENT_UNUSED)
	    break;
    }
    FAT12_seek (fp, -DIRENT_SIZE, SEEK_CUR);
    dirent->entry.clus_num    = fp->current_clus_num;
    dirent->entry.clus_offset = fp->current_clus_offset;
    encode_dirent (ptr, dirent);
    ret = FAT12_write (fp, ptr, DIRENT_SIZE);

    /* write EndOfDir */
    if (!is_end_of_directory (dirent) && is_end_of_directory (&dirent_tmp)) {
	dirent_tmp = *dirent;
	dirent_tmp.name [0] = DIRENT_END_DIR;
	encode_dirent (ptr, &dirent_tmp);
	ret = FAT12_write (fp, ptr, DIRENT_SIZE);
	FAT12_seek (fp, -DIRENT_SIZE, SEEK_CUR);
    }
    return ret;
}

int
FAT12_seek   (struct file *fp, int offset, int whence)
{
    switch (whence) {
    case SEEK_SET: fp->current_offset = offset;  break;
    case SEEK_CUR: fp->current_offset += offset; break;
    case SEEK_END: fp->current_offset = get_filesize (&fp->dirent) + offset;
	                                         break;
    default:       return -1;
    }
    if (fp->current_offset < 0) {
	printf ("FAT12_seek: negative offset: %d\n", fp->current_offset);
	fp->current_offset = 0;
    }
    calc_clus_num_offset (fp, &fp->current_clus_num, &fp->current_clus_offset);
    return 0;
}

int
FAT12_mkdir  (const char *path)
{
    struct file file_dot, file_dotdot;
    struct FAT_dirent dirent;
    int clus_num, ret;
    char *dir, file [FILENAME_LEN + 1];
    uint8_t ptr [DIRENT_SIZE];

    ret = FAT12_open (&file_dot, path, O_RDONLY);
    if (ret != -1) {
	printf ("FAT12_mkdir: already exists: %s\n", path);
	return -1;
    }
    FAT12_close (&file_dot);

    clus_num = alloc_cluster (FIRST_CLUSTER);
    ret = create_dirent (path, &dirent, ATTR_DIRECTORY, clus_num);
    if (ret == -1)
	return -1;

    ret = FAT12_open (&file_dot, path, O_WRONLY);
    assert (ret != -1);

    /* create "." entry */
    dirent = file_dot.dirent;
    memcpy (dirent.name, DIR_DOT, FILENAME_LEN + 1);
    encode_dirent (ptr, &dirent);
    FAT12_write (&file_dot, ptr, DIRENT_SIZE);

    /* create ".." entry */
    dir = divide_path_right (path, file);
    FAT12_open (&file_dotdot, dir, O_RDONLY);
    if (file_dotdot.dirent.clus_num == ROOT_DIR_CLUS_NUM) { 
	memcpy (&dirent, &FAT_info.DummyRootDirent, DIRENT_SIZE);
    } else {
	FAT12_readdir (&file_dotdot, &dirent, 0);
	assert (!strcmp (dirent.name, DIR_DOT));
    }
    FAT12_close (&file_dotdot);
    memcpy (dirent.name, DIR_DOTDOT, FILENAME_LEN + 1);
    encode_dirent (ptr, &dirent);
    FAT12_write (&file_dot, ptr, DIRENT_SIZE);

    /* create end-of-dir entry, first */
    dirent.name [0] = DIRENT_END_DIR;
    encode_dirent (ptr, &dirent);
    FAT12_write (&file_dot, ptr, DIRENT_SIZE);

    FAT12_close (&file_dot);
    return 0;
}

int
FAT12_rmdir  (const char *path)
{
    struct file file_dot;
    struct FAT_dirent dirent;
    int ret, n_files;

    ret = FAT12_open (&file_dot, path, O_RDONLY);
    if (ret == -1) {
	printf ("FAT12_rmdir: No such directory: %s\n", path);
	return -1;
    }
    if (!is_directory (&file_dot.dirent)) {
	printf ("FAT12_rmdir: No directory: %s\n", path);
	return -1;
    }

    n_files = 0;
    while (FAT12_readdir (&file_dot, &dirent, 1) > 0)
	n_files++;
    if (n_files > 2) {
	printf ("FAT12_rmdir: Directory not empty: %s\n", path);
	return -1;
    }
    FAT12_close (&file_dot);

    FAT12_unlink (path);
    return 0;
}

int
FAT12_unlink (const char *path)
{
    int ret;
    struct FAT_dirent dirent;
    ret = path_to_dirent (path, &dirent);
    if (ret == -1) {
	printf ("FAT12_unlink: No such file: %s\n", path);
	return -1;
    }
    free_cluster_chain (dirent.clus_num);
    dirent.clus_num = CLUSTER_FREE;
    dirent.name [0] = DIRENT_UNUSED;
    write_dirent (dirent.entry.clus_num, dirent.entry.clus_offset, &dirent);
    return 0;
}

int
FAT12_stat   (const char *path, struct FAT_dirent *dirent)
{
    return path_to_dirent (path, dirent);
}

#if !defined(KERNEL)
int
main (int argc, char *argv [])
{
    int i;

    if (argc < 2) {
    	fprintf (stderr, "Usage(dump) :%s fd_image\n", argv [0]);
    	fprintf (stderr, "Usage(add)  :%s fd_image -write file...\n", argv [0]);
    	fprintf (stderr, "Usage(rm)   :%s fd_image -unlink file...\n", argv [0]);
    	fprintf (stderr, "Usage(mkdir):%s fd_image -mkdir dir...\n", argv [0]);
    	fprintf (stderr, "Usage(rmdir):%s fd_image -rmdir dir...\n", argv [0]);
	exit (1);
    }

    if ((image_fp = fopen (argv [1], "rb+")) == NULL) {
	fprintf (stderr, "No such image file: %s\n", argv [1]);
	exit (1);
    }

    FAT12_init ();

    if (argc == 2) {
	is_dump = 1;
	FAT12_dump ();
	return 0;
    }

    if (!strcmp (argv [2], "-write"))
	for (i = 3; i < argc; i++)
	    FAT12_add (argv [i]);
    else if (!strcmp (argv [2], "-mkdir"))
	for (i = 3; i < argc; i++)
	    FAT12_mkdir (argv [i]);
    else if (!strcmp (argv [2], "-rmdir"))
	for (i = 3; i < argc; i++)
	    FAT12_rmdir (argv [i]);
    else if (!strcmp (argv [2], "-unlink"))
	for (i = 3; i < argc; i++)
	    FAT12_unlink (argv [i]);

    FAT12_sync ();
    fclose (image_fp);
    return 0;
}
#endif /* !KERNEL */
/* ---------------------------------------------------------------------- */
