/*
    udos: micro disk OS.
    Copyright (c) 2003, 2004   Katsuhiko Gondow (gondow@cs.titech.ac.jp)
    $Id: fd.c,v 1.27 2004/07/14 05:43:19 gondow Exp gondow $
*/
/*
  See also the programming guide (p.39 and later in 82077.pdf).
 */
#include "misc.h" /* for int8_t, panic, assert, printk */
#include "vga.h"  /* for inb, outb, delay, PIT_ticks */
#include "fd.h"
#include "paging.h" /* for kmalloc */
#include "list.h"
#include "task.h" /* for sync_wait/sync_notify */
/* ---------------------------------------------------------------------- */
enum {
    /* reg. addresses for FDC#1 (udos does not use FDC#2) */
    FDC_DOR = 0x3F2,         /* digital output reg. */
    FDC_MSR = FDC_DOR + 2,   /* main status reg. */
    FDC_DR  = FDC_DOR + 3,   /* data reg. (aka FIFO) */
    FDC_DIR = FDC_DOR + 5,   /* data input reg. */
    FDC_CCR = FDC_DOR + 5,   /* config. control reg. */

    FDC_INTR_TIMEOUT   = 5000, /* in msec */
     
    /* parameters for 3.5inch 1.44MB MFM floppy disk */
    FD_SECTOR_SIZE     = 512,
    HEADS_PER_CYLINDER = 2,  /* number of heads per cylinder */
    SECTORS_PER_TRACK  = 18, /* number of sectors per track */
    CYLINDERS_PER_DISK = 80, /* number of cylinders per disk */
    FD_N               = 2,  /* sector size (128*2^N byte/sector) */
    FD_EOT             = 18, /* end of track */
    FD_GSL             = 0x1B, /* gap length */
    FD_GPL             = 0x6C, /* gap skip length */
    FD_DTL             = 0xFF, /* data length (0xFF is meaningless) */

    MAX_RETRY = 0x1000,        /* xxx */

    /* reg. addresses for DMAC, DMA2(slave) channel */
    DMAC_status_reg       = 0x08,
    DMAC_instruction_reg  = DMAC_status_reg,
    DMAC_request_reg      = DMAC_status_reg + 1,
    DMAC_channel_mask_reg = DMAC_status_reg + 2,
    DMAC_mode_reg         = DMAC_status_reg + 3,
    DMAC_clear_byte_point_flipflop = DMAC_status_reg + 4,
    DMAC_interim_reg      = DMAC_status_reg + 5,
    DMAC_master_clear     = DMAC_status_reg + 6,
    DMAC_mask_reg         = DMAC_status_reg + 7,

    DMAC_page_reg_2       = 0x81,
    DMAC_curr_addr_reg_2  = 0x04,
    DMAC_base_addr_reg_2  = DMAC_curr_addr_reg_2,
    DMAC_curr_count_reg_2 = 0x05,
    DMAC_base_count_reg_2 = DMAC_curr_count_reg_2,
};

enum {
    MAX_BUFFER_NUM        = 32,

    BUFFER_IS_DIRTY       = 0x00000001,
    BUFFER_IS_LOCKED      = 0x00000002,
};

struct buffer_header {
    int block_num;   /* in LBA, -1 means unuse */
    uint32_t status; /* combination of BUFFER_* */
    void *buffer;
};
/* ---------------------------------------------------------------------- */
volatile int motor_off_time = 0;
/* ---------------------------------------------------------------------- */
static struct CDLL buffer_list;
static volatile int fd_is_busy = 0;
static int motor_on   = 0;
static int ident_fd;
/* ---------------------------------------------------------------------- */
static void    FDC_busy_sleep  (int usec);
static void    FDC_block_sleep (int msec, int needs_interrupt);
static void    FDC_send_byte (uint8_t byte);
static uint8_t FDC_read_byte (void);
static int     FDC_CHS2LBA (int cylinder, int head, int sector);
static void    FDC_LBA2CHS (int lba, int *cylinder, int *head, int *sector);

static void    FDC_recalibrate (int head);
static void    FDC_seek  (int cylinder, int head, int sector);
static uint8_t FDC_sense_interrupt_status (void);
static void    FDC_setup_DMAC (uint32_t physical_addr, int count, int is_read);
static void    FDC_read_write (int lba, int block_count, void *paddr,
			       int is_read);

static struct CDLL *search_buffer (int block_num);
static struct CDLL *search_free_buffer (void);
/* ---------------------------------------------------------------------- */
static void
FDC_busy_sleep (int usec)
{
    int i;
    for (i = 0; i < usec; i++)
	delay (); /* busy waiting */
}

static void
FDC_block_sleep (int msec, int needs_interrupt)
{
    int ret;
    ret = sync_wait (&ident_fd, msec, UNINTERRUPTIBLE);
    if (needs_interrupt && ret == -1) {
	printk ("pid=%d, msec=%d\n", sys_getpid (), msec);
	panic ("no interrupt");
    }
}

static void
FDC_send_byte (uint8_t byte)
{
    int i;
    uint8_t status;
    for (i = 0; i < MAX_RETRY; i++) {
	status = inb (FDC_MSR);
	if ((status & 0xC0) == 0x80) { /* MSR == 10XXXXXX */
	    outb (byte, FDC_DR);
	    return;
	}
	FDC_busy_sleep (10);
    }
    printk ("MSR=%x\n", status);
    panic ("FDC_send_byte");
}

static uint8_t
FDC_read_byte (void)
{
    int i;
    uint8_t status;
    for (i = 0; i < MAX_RETRY; i++) {
	status = inb (FDC_MSR);
	if ((status & 0xC0) == 0xC0) { /* MSR == 11XXXXXX */
	    return inb (FDC_DR);
	}
	FDC_busy_sleep (10);
    }
    printk ("MSR=%x\n", status);
    panic ("FDC_read_byte");
    return -1; /* not reached here */
}

static int
FDC_CHS2LBA (int cylinder, int head, int sector)
{
    int lba = (cylinder * HEADS_PER_CYLINDER + head) * SECTORS_PER_TRACK
	+ sector - 1;
    assert (0 <= cylinder && cylinder < CYLINDERS_PER_DISK);
    assert (0 <= head     && head    < HEADS_PER_CYLINDER);
    assert (0 <  sector   && sector  <= SECTORS_PER_TRACK); /* 1-origin */
    return lba;
}

static void
FDC_LBA2CHS (int lba, int *cylinder, int *head, int *sector)
{
    int sectors_per_cylinder = HEADS_PER_CYLINDER * SECTORS_PER_TRACK;
    int tmp  = lba % sectors_per_cylinder;
    if (!(0 <= lba && lba < (CYLINDERS_PER_DISK * HEADS_PER_CYLINDER
			     * SECTORS_PER_TRACK))) {
	printk ("lba=%d, %x\n", lba, lba);
    }
    assert (0 <= lba && lba < (CYLINDERS_PER_DISK * HEADS_PER_CYLINDER
			       * SECTORS_PER_TRACK));
    *cylinder = lba / sectors_per_cylinder;
    *head     = tmp / SECTORS_PER_TRACK;
    *sector   = tmp % SECTORS_PER_TRACK + 1;
}

static uint8_t
FDC_sense_interrupt_status (void)
{
    int i;
    uint8_t st0, pcn;
    for (i = 0; i < MAX_RETRY; i++) {
	FDC_send_byte (0x08); /* issue SENSE-INTERRUPT-STATUS command */
	st0 = FDC_read_byte ();
	if (st0 != 0x80)
	    goto success;
    }
    panic ("cannot execute SENSE-INTERRUPT-STATUS command");

 success:
    pcn = FDC_read_byte ();
    return st0;
}

static void
FDC_recalibrate (int head)
{
    FDC_send_byte (0x07); /* issue RECALIBRATE command */
    FDC_send_byte (head << 2); /* HD=head, US1:US0=0:0 */
}

/* before calling FDC_seek, FDC_motor (1) needs to be called */
static void
FDC_seek (int cylinder, int head, int sector)
{
    FDC_send_byte (0x0F); /* issue SEEK command */
    FDC_send_byte (head << 2); /* HD=head, US1:US0=0:0 */
    FDC_send_byte (cylinder); /* NCN=cylinder */
}

static void
FDC_setup_DMAC (uint32_t physical_addr, int count, int is_read)
{
    /* disable DMA2 (mask set for DMA2) */
    outb (0x06, DMAC_channel_mask_reg); /* STCL=1, SEL1:SEL0=2 */

    /* set address and counter */
    outb (0xFF, DMAC_clear_byte_point_flipflop); /* 0xFF is meaningless */
    outb (physical_addr       & 0xFF, DMAC_base_addr_reg_2);
    outb (physical_addr >> 8  & 0xFF, DMAC_base_addr_reg_2);
    outb (physical_addr >> 16 & 0xFF, DMAC_page_reg_2);
    count--;
    outb (count      & 0xFF, DMAC_base_count_reg_2);
    outb (count >> 8 & 0xFF, DMAC_base_count_reg_2);

    /* set mode */
    if (is_read) {
	/* single, inc, noauto, write (I/O device -> memory), DMA2 */
	outb (0x46, DMAC_mode_reg);
    } else {
	/* single, inc, noauto, read (memory -> I/O device), DMA2 */
	outb (0x4A, DMAC_mode_reg);
    }

    /* enable DMA2 (mask clear for DMA2) */
    outb (0x02, DMAC_channel_mask_reg); /*STCL=0, SEL1:SEL0=2 */
}

static void
FDC_read_write (int lba, int block_count, void *paddr, int is_read)
{
    int cylinder, head, sector;
    uint8_t st0, st1, st2, c, h, r, n;
    uint32_t eflags;

    if (block_count == 0)
	return;

    klock2 (eflags);
    while (fd_is_busy) {
	sync_wait ((void *)&fd_is_busy, 0, INTERRUPTIBLE);
	process_signals (1);
    }
    fd_is_busy = 1;
    kunlock2 (eflags);

    FDC_LBA2CHS (lba, &cylinder, &head, &sector);
#if 0 /* for debug */
    printk ("lba=%d, cyl=%d, head=%d, sect=%d\n", lba, cylinder, head, sector);
#endif
    assert (0 < block_count && block_count <= 128); /* must be < 64KB */

    klock ();
    motor_off_time = 0; /* cancel automatic motor off */
    kunlock ();
    if (!motor_on) {
	FDC_motor (1); /* motor on */
	outb (0x00, FDC_CCR); /* set data rate to 500kbps */
	klock2 (eflags);
	FDC_block_sleep (500, 0); /* 500ms for motor ready (p.44, 82077.pdf)*/
	kunlock2 (eflags);
    }

    klock2 (eflags);
    FDC_recalibrate (head);
    FDC_block_sleep (FDC_INTR_TIMEOUT, 1); /* wait for interrupt */
    kunlock2 (eflags);
    st0 = FDC_sense_interrupt_status ();
    assert (st0 == 0x20 || st0 == 0x24); /* bochs-2.1.1 issues 0x24 */

    klock2 (eflags);
    FDC_seek (cylinder, head, sector);
    FDC_block_sleep (FDC_INTR_TIMEOUT, 1); /* wait for interrupt */
    kunlock2 (eflags);
    st0 = FDC_sense_interrupt_status ();
    assert (st0 == 0x20 || st0 == 0x24); /* bochs-2.1.1 issues 0x24 */

    FDC_setup_DMAC ((uint32_t)paddr, block_count * FD_SECTOR_SIZE, is_read);

    if (is_read)
	FDC_send_byte (0xE6); /* issue READ command (MT=MF=SK=1) */
    else
	FDC_send_byte (0xC5); /* issue WRITE command (MT=MF=1, SK=0) */

    FDC_send_byte (head << 2);   /* HD=head, US1:US0=0:0 */
    FDC_send_byte (cylinder);    /* C */
    FDC_send_byte (head);        /* H */
    FDC_send_byte (sector);      /* R */
    FDC_send_byte (FD_N);        /* N */
    FDC_send_byte (FD_EOT);      /* EOT */
    FDC_send_byte (FD_GSL);      /* GSL */
    klock2 (eflags);
    FDC_send_byte (FD_DTL);      /* DTL */
    FDC_block_sleep (FDC_INTR_TIMEOUT, 1); /* wait for interrupt */
    kunlock2 (eflags);

    st0 = FDC_read_byte ();
    st1 = FDC_read_byte ();
    st2 = FDC_read_byte ();
    c   = FDC_read_byte ();
    h   = FDC_read_byte ();
    r   = FDC_read_byte ();
    n   = FDC_read_byte ();
#if 0 /* for debug */
    printk ("st0=0x%x, st1=0x%x, st2=0x%x, c=%d, h=%d, r=%d, n=%d\n",
	    st0, st1, st2, c, h, r, n);
#endif
    assert ((st0 & 0xC0) == 0);

    /* set delayed motor off time */
    klock ();
    motor_off_time = PIT_ticks + (1000 / PIT_MS); /* 1 sec delay */
    kunlock ();

    klock2 (eflags);
    fd_is_busy = 0;
    sync_notify ((void *)&fd_is_busy);
    kunlock2 (eflags);
}

static struct CDLL *
search_buffer (int block_num)
{
    struct CDLL *cdll;
    for (cdll = buffer_list.next; cdll != &buffer_list; cdll = cdll->next) {
	struct buffer_header *header = (struct buffer_header *)cdll->data;
	if (header->block_num == block_num) {
	    return cdll;
	}
    }
    return NULL;
}

static struct CDLL *
search_free_buffer (void)
{
    struct CDLL *cdll;
    struct buffer_header *header;

    for (cdll = buffer_list.next; cdll != &buffer_list; cdll = cdll->next) {
	header = (struct buffer_header *)cdll->data;
	if (!(header->status & BUFFER_IS_LOCKED))
	    break;
    }
    if (cdll == &buffer_list) {
	panic ("no free buffer (should wait here)");
    }
    if (header->status & BUFFER_IS_DIRTY) {
#if 1
	printk ("search_free_buffer: #%d writing back...\n",
		header->block_num);
#endif
	FDC_write (header->block_num, 1, header->buffer); /* write back */
    }
    header->block_num = -1;
    header->status = 0;
    return cdll;
}
/* ---------------------------------------------------------------------- */
void
FDC_motor (int on)
{
    /* motor#0 on/off, DMA on, #RESET off, select drive#0 */
    outb ((on ? 0x1C : 0x0C), FDC_DOR);
    motor_on = (on ? 1 : 0);
}

void
buffer_init (void)
{
    int i;

    CDLL_wrap (&buffer_list, NULL);

    /* initialize buffer-cache */
    for (i = 0; i < MAX_BUFFER_NUM; i++) {
	struct buffer_header *header = kmalloc (sizeof (struct buffer_header));
	struct CDLL *cdll = CDLL_create (header);
	CDLL_insert_head (&buffer_list, cdll);

	header->block_num = -1;
	header->status    = 0;
	header->buffer    = kmalloc (FD_SECTOR_SIZE);
    }
}

void
FDC_intr_handler   (void)
{
    outb (0x20, 0x20); /* send EOI to PIC1 */
    sync_notify (&ident_fd);
}

void
FDC_init (void)
{
    uint32_t eflags;
    int i;

    klock2(eflags);
    outb (0x00, FDC_DOR); /* clear bit2 (start #RESET) */
    FDC_busy_sleep (10);  /* 7.1us is required for resetting 82077A */
    outb (0x04, FDC_DOR); /* set   bit2 (end #RESET) */
    outb (0x00, FDC_CCR); /* set data rate to 500kbps */
    FDC_block_sleep (FDC_INTR_TIMEOUT, 1); /* wait for interrupt */
    kunlock2(eflags);

    for (i = 0; i < 4; i++) {
	uint8_t st0 = FDC_sense_interrupt_status ();
	assert (st0 == (0xC0 | (uint8_t)i));
    }

    /* issue SPECIFY command */
    FDC_send_byte (0x03); /* SPECIFY command */
    FDC_send_byte ((0x0C << 4) | 0x01); /* SRT=0x0C(4ms), HUT=1 */
    FDC_send_byte ((0x08 << 1) | 0x00); /* HLT=0x08(16ms), ND=0(DMA) */
}

void
FDC_read (int lba, int block_count, void *paddr)
{
    FDC_read_write (lba, block_count, paddr, 1);
}

void
FDC_write (int lba, int block_count, void *paddr)
{
    FDC_read_write (lba, block_count, paddr, 0);
}

void
buffer_read  (int lba, void *paddr)
{
    struct CDLL *cdll = search_buffer (lba);
    struct buffer_header *header;
#if 0
    printk ("buffer_read: lba=%x, paddr=%x\n", lba, paddr);
#endif
    if (cdll == NULL) {
	cdll = search_free_buffer ();
	header = (struct buffer_header *)cdll->data;
	FDC_read (lba, 1, header->buffer);
	header->block_num = lba;
    } else {
	header = (struct buffer_header *)cdll->data;
    }
    memcpy (paddr, header->buffer, FD_SECTOR_SIZE);

    /* LRU rotating */
    CDLL_remove (cdll);
    CDLL_insert_tail (&buffer_list, cdll);
}

void
buffer_readn (int lba, int block_count, void *paddr)
{
    int i, offset, is_all_cached = 1;
    struct CDLL *cdll;
    struct buffer_header *header;
    char *buf = kmalloc (block_count * FD_SECTOR_SIZE);

#if 0
    printk ("lba=%x, block_count=%x, paddr=%x\n", lba, block_count, paddr);
#endif
    for (i = 0; i < block_count; i++) {
	cdll = search_buffer (lba + i);
	if (cdll == NULL) {
	    is_all_cached = 0;
	    break;
	}
    }

    if (!is_all_cached) {
	FDC_read (lba, block_count, buf);
    }
    for (i = 0; i < block_count; i++) {
	offset = FD_SECTOR_SIZE * i;
	cdll = search_buffer (lba + i);
	if (cdll != NULL) {
	    buffer_read (lba + i, (char *)paddr + offset);
	    continue;
	} 
	cdll   = search_free_buffer ();
	header = (struct buffer_header *)cdll->data;
	memcpy (header->buffer, buf + offset, FD_SECTOR_SIZE);
	memcpy ((char *)paddr + offset, header->buffer, FD_SECTOR_SIZE);
	header->block_num = lba + i;
	header->status    = 0;
	CDLL_remove (cdll);
	CDLL_insert_tail (&buffer_list, cdll);
    }
    kfree (buf);
}

void
buffer_write  (int lba, void *paddr)
{
    struct CDLL *cdll = search_buffer (lba);
    struct buffer_header *header;
#if 0
    printk ("lba=%x\n, paddr=%x\n", lba, paddr);
#endif
    if (cdll == NULL) {
	cdll = search_free_buffer ();
	header = (struct buffer_header *)cdll->data;
	header->block_num = lba;
    } else {
	header = (struct buffer_header *)cdll->data;
    }
    memcpy (header->buffer, paddr, FD_SECTOR_SIZE);
    header->status |= BUFFER_IS_DIRTY;

    /* LRU rotating */
    CDLL_remove (cdll);
    CDLL_insert_tail (&buffer_list, cdll);
}

void
buffer_writen (int lba, int block_count, void *paddr)
{
    int i;
#if 0
    printk ("lba=%x, block_count=%x, paddr=%x\n", lba, block_count, paddr);
#endif
    for (i = 0; i < block_count; i++) {
	buffer_write (lba + i, (char *)paddr + FD_SECTOR_SIZE * i);
    }
}

void
buffer_sync   (void)
{
    struct CDLL *cdll;
    for (cdll = buffer_list.next; cdll != &buffer_list; cdll = cdll->next) {
	struct buffer_header *header = (struct buffer_header *)cdll->data;
	if (header->block_num == -1) continue;
	if (header->status & BUFFER_IS_DIRTY) {
#if 1
	    printk ("buffer_sync: block #%x writing back...\n",
		    header->block_num);
#endif	    
	    FDC_write (header->block_num, 1, header->buffer);
	}
	header->status &= ~BUFFER_IS_DIRTY;
    }
}
/* ---------------------------------------------------------------------- */
