/*
    udos: micro disk OS.
    Copyright (c) 2003, 2004   Katsuhiko Gondow (gondow@cs.titech.ac.jp)
    $Id: paging.c,v 1.33 2004/07/14 05:43:19 gondow Exp gondow $
*/
#include "vga.h"
#include "paging.h"
#include "misc.h"
#include "boot32.h"
#include "list.h"
#include "fat12.h"
#include "task.h"

/*
  0MB -  4MB: identity mapping
  1MB -  3MB: managed by kmalloc/kfree
  3MB -  4MB: managed by im_page_alloc/im_page_free
  4MB -  max: managed by page_alloc/page_free
 */
enum {
    /* for identity-mapping page allocator */
    IM_MEM_SIZE        = 1024 * 1024,            /*  1MB */
    IM_PAGE_BASE_PADDR = 0x300000,               /*  3MB */
    IM_PAGE_NUM        = IM_MEM_SIZE / PAGE_SIZE,

    MEM_SIZE           = 12 * 1024 * 1024,    	 /* 12MB */
    PAGE_BASE_PADDR    = 0x400000,               /*  4MB */
    PAGE_NUM           = MEM_SIZE / PAGE_SIZE,

    PAGE_TBL_MAP_ADDR  = 0x400000,
    PAGE_TBL_MAP_ADDR2 = 0x800000,
    PAGE_MAP_ADDR2     = 0xC00000,
};

enum {
    PTE_P    = 0x001,	   /* present */
    PTE_RW   = 0x002,      /* read/write */
    PTE_US   = 0x004,      /* user/superviser */
    PTE_PWT  = 0x008,      /* page write through/back */
    PTE_PCD  = 0x010,      /* page cache disable */
    PTE_A    = 0x020,      /* access */
    PTE_D    = 0x040,      /* dirty */
    PTE_PS   = 0x080,      /* page size */
    PTE_G    = 0x100,      /* global page */
};

#define ROUNDUP(p, n)    ((((uint32_t)p)+((n)-1)) & ~((n)-1))
#define ROUNDUP_ALIGN(p) ROUNDUP(p, KMALLOC_ALIGNMENT)

#define FLUSH_TBL        asm volatile ("movl %cr3, %eax; movl %eax, %cr3")

struct kmalloc_header {
    uint32_t magic; /* shoud be equal to KMALLOC_MAGIC */
    uint32_t size;  /* excluding header size */
    int      is_free;
};

enum {
    KMALLOC_MAGIC         = 0xDEADBEAF,
    KMALLOC_ALIGNMENT     = 4, /* see p.28 in abi386-4.pdf */
    KMALLOC_HEADER_SIZE1  = ROUNDUP_ALIGN (sizeof (struct CDLL)),
    KMALLOC_HEADER_SIZE2  = ROUNDUP_ALIGN (sizeof (struct kmalloc_header)),
    KMALLOC_HEADER_SIZE12 = KMALLOC_HEADER_SIZE1 + KMALLOC_HEADER_SIZE2,
    KMALLOC_ADDR_START    = 0x00100000, /* 1MB */
    KMALLOC_ADDR_END      = 0x00300000, /* 4MB */
};

/* ---------------------------------------------------------------------- */
static uint32_t used_page_bitmap [1 + PAGE_NUM / 32]; /* 1 if used */
static uint32_t used_im_page_bitmap [1 + IM_PAGE_NUM / 32]; /* 1 if used */
static struct CDLL kmalloc_list;
/* ---------------------------------------------------------------------- */
static uint32_t make_PTE     (void *paddr, uint32_t pte_us);
/* ---------------------------------------------------------------------- */
static uint32_t
make_PTE (void *paddr, uint32_t pte_us)
{
    return ((uint32_t)paddr & 0xFFFFF000) | PTE_P | PTE_RW | pte_us;
}

static struct kmalloc_header *
divide_block (struct CDLL *cdll1, uint32_t size)
{
    struct kmalloc_header *header1 = cdll1->data;
    struct CDLL *cdll2 = (void*)((char *)cdll1 + KMALLOC_HEADER_SIZE12 + size);
    struct kmalloc_header *header2 = (void*)((char *)cdll2 + KMALLOC_HEADER_SIZE1);

    header2->size = header1->size - (KMALLOC_HEADER_SIZE12 + size);
    header2->magic = KMALLOC_MAGIC;
    header2->is_free = 1;

    header1->size = size;
    assert (header1->magic == KMALLOC_MAGIC);
    assert (header1->is_free == 1);

    CDLL_wrap (cdll2, header2);
    CDLL_insert_head (cdll1, cdll2);

    return header1;
}

static void
merge_block (struct CDLL *cdll)
{
    struct kmalloc_header *header1 = cdll->data;
    struct kmalloc_header *header2 = cdll->next->data;
    if (cdll == &kmalloc_list || cdll->next == &kmalloc_list)
	return;
    if (!header1->is_free || !header2->is_free)
	return;
    header1->size += header2->size + KMALLOC_HEADER_SIZE12;
    header2->magic = 0; /* nullify magic number */
    CDLL_remove (cdll->next);
}
/* ---------------------------------------------------------------------- */
void
setup_paging (void)
{
    int i;

    /* clear all page directory entries */
    memset (kernel_page_dir, 0, PAGE_SIZE);

    /* initialize a page table for identity-mapping */
    kernel_page_dir [0] = make_PTE (kernel_page_tbl_IM, 0);

    for (i = 0; i < 1024; i++) {
	/* register 0-4MB with paddr==vaddr. */
	kernel_page_tbl_IM [i] = make_PTE ((uint32_t *)(i * 0x1000), 0);
    }

    kernel_page_dir [1] = make_PTE (kernel_page_dir, 0);

    /* set the page_dir address into %cr3 */
    asm volatile ("movl  %0, %%cr3"::"r"(kernel_page_dir));

    /* setting page-fault handler to IDT is already done in cpu.c */

    /* set %cr0's PG bit (paging on). 
       From now on, any physical addresses cannot be dereferenced. */
    asm volatile ("movl %cr0, %eax\n"
		  "orl  0x80000000, %eax\n"
		  "movl %eax, %cr0");
}

/* Create and initialize a user address space. */
void
create_aspace (uint32_t *page_dir, char *filename, uint32_t size)
{
    struct file file;
    char buf [512];
    int  nbytes;
    void *p;

    assert (0); /* obsolete */
    uint32_t code_page_vaddr  = USER_CODE_ADDR;
    uint32_t code_dir_idx     = code_page_vaddr >> 22;

    uint32_t stack_page_vaddr = USER_STACK_ADDR - 16;
    uint32_t stack_dir_idx    = stack_page_vaddr >> 22;

    page_dir [0] = kernel_page_dir [0]; /* copy PTE for IM area */
    page_dir [1] = kernel_page_dir [1]; /* copy PTE for PTM area */

    /* setup user code area */
    p = (void *)code_page_vaddr;
    FAT12_open (&file, filename, O_RDONLY);
    while ((nbytes = FAT12_read (&file, buf, sizeof (buf))) > 0) {
	memcpy (p, buf, nbytes);
	p += nbytes;
    }
    FAT12_close (&file);

    page_dir [code_dir_idx] = kernel_page_dir [code_dir_idx];
    kernel_page_dir [code_dir_idx] = 0;
    /* xxx: this causes 4MB memory moving. $B$3$l$b$+$C$30-$$(B */

    /* setup user stack area, now just allocating only 4KB */
    kmem_map (stack_page_vaddr);
    page_dir [stack_dir_idx] = kernel_page_dir [stack_dir_idx];
    kernel_page_dir [stack_dir_idx] = 0;
}

void
page_fault_handler (uint32_t error)
{
    uint32_t cr2;

#if 0 /* xxx */
    asm volatile ("pushw %ds; movw $0x10, %ax; movw %ax, %ds; subl $32, %esp");
#endif
    asm volatile ("movl %%cr2,    %0":"=r"(cr2));

    if (error & 1) {
#if 0
	printk ("page_fault_handler: pid=%d, error=%x, cr2=%x\n",
		sys_getpid (), error, cr2);
#endif
	sys_kill (sys_getpid (), SIGSEGV);
	process_signals (1);

	/* never reached here */
	panic ("page_fault_handler: page level protection error");
    }

#if 0 /* for debug */
    printk ("page fault: cr2=%x (pid = %d)\n", cr2, sys_getpid ());
#endif

    kmem_map (cr2);
#if 0 /* xxx */
    asm volatile ("addl $32, %esp; pop %ds");
#endif
}

/* mark a free page as used, and return its physical address */
void *
page_alloc (void)
{
    int i, j;
    
    for (i = 0; i < PAGE_NUM / 32; i++) {
	if (used_page_bitmap [i] != 0xFFFFFFFF) {
	    for (j = 0; j < 32; j++) {
		if (((1 << j) & used_page_bitmap [i]) == 0) {
		    used_page_bitmap [i] |= (1 << j);
		    return (void *)((i * 32 + j)* PAGE_SIZE + PAGE_BASE_PADDR);
		}
	    }
	    panic ("page_alloc: error\n");
	}
    }
    panic ("page_alloc: no page\n");
    return 0;
}

/* mark a given page as free */
void
page_free  (void *page)
{
    int i, j;
    uint32_t page_num;

    page_num = ((uint32_t)page - PAGE_BASE_PADDR) / PAGE_SIZE;
    i = page_num / 32;
    j = page_num % 32;

    if (!(used_page_bitmap [i] & (1 << j))){
	printk ("page_free: page=%x\n", page);
	panic  ("page_free: error");
    }
    used_page_bitmap [i] &= ~(1 << j);
}

void *
im_page_alloc (void)
{
    int i, j;
    
    for (i = 0; i < IM_PAGE_NUM / 32; i++) {
	if (used_im_page_bitmap [i] != 0xFFFFFFFF) {
	    for (j = 0; j < 32; j++) {
		if (((1 << j) & used_im_page_bitmap [i]) == 0) {
		    used_im_page_bitmap [i] |= (1 << j);
		    return (void *)((i * 32 + j)* PAGE_SIZE
				    + IM_PAGE_BASE_PADDR);
		}
	    }
	    panic ("im_page_alloc: error\n");
	}
    }
    panic ("im_page_alloc: no im page\n");
    return 0;
}

void
im_page_free  (void *page)
{
    int i, j;
    uint32_t page_num;

    page_num = ((uint32_t)page - IM_PAGE_BASE_PADDR) / PAGE_SIZE;
    i = page_num / 32;
    j = page_num % 32;

    if (!(used_im_page_bitmap [i] & (1 << j))){
	panic ("im_page_free: error");
    }
    used_im_page_bitmap [i] &= ~(1 << j);
}

void
setup_kmalloc (void)
{
    struct CDLL *all_kernel_heap;
    struct kmalloc_header *header;

    CDLL_wrap (&kmalloc_list, NULL);
    header = (struct kmalloc_header*)(KMALLOC_ADDR_START
				      + KMALLOC_HEADER_SIZE1);
    all_kernel_heap = (struct CDLL *)KMALLOC_ADDR_START;
    CDLL_wrap (all_kernel_heap, header);
    CDLL_insert_head (&kmalloc_list, all_kernel_heap);
    header->magic = KMALLOC_MAGIC;
    header->size  = KMALLOC_ADDR_END - KMALLOC_ADDR_START
		    - KMALLOC_HEADER_SIZE12;
    header->is_free = 1;
}

void *
kmalloc (uint32_t size)
{
    struct CDLL *p;
    void *ret;

    size = ROUNDUP_ALIGN (size);
    for (p = kmalloc_list.next; p != &kmalloc_list; p = p->next) {
	struct kmalloc_header *header = p->data;
	/* first-fit simple allocater */
	if (header->is_free && header->size >= size) {
	    if (header->size >= size + KMALLOC_HEADER_SIZE12) {
		header = divide_block (p, size);
	    } 
	    header->is_free = 0;
	    ret = (char *)header + KMALLOC_HEADER_SIZE2;
#if defined(DEBUG)
	    printk ("kmalloc: ptr=%x, size=%d\n", ret, size);
#endif
	    return ret;
	}
    }
    panic ("kmalloc: no memory\n");
    return NULL;
}

void
kfree (void *ptr)
{
    struct CDLL *cdll = (struct CDLL *)((char *)ptr - KMALLOC_HEADER_SIZE12);
    struct kmalloc_header *header = cdll->data;
#if defined(DEBUG)
    printk ("kfree: ptr=%x\n", ptr);
#endif
    assert (header->magic == KMALLOC_MAGIC);
    assert (!header->is_free);
    header->is_free = 1;
    merge_block (cdll);
    merge_block (cdll->prev);
}

void
kmem_map  (uint32_t vaddr)
{
    uint32_t *page_dir, *page_tbl_vaddr, *pte_vaddr;
    uint32_t page_dir_index = vaddr >> 22;          /* upper  10 bits */
    uint32_t page_tbl_index = (vaddr>> 12) & 0x3FF; /* middle 10 bits */
#if 0
    uint32_t page_offset    = vaddr & 0xFFF;        /* lower  12 bits */
#endif
    
    asm volatile ("movl %%cr3,    %0":"=r"(page_dir));
    page_tbl_vaddr = (uint32_t *)(PAGE_TBL_MAP_ADDR + (page_dir_index << 12));

    /* if no page table is allocated for 'vaddr', allocate it */
    if (!(page_dir [page_dir_index] & PTE_P)) {
	uint32_t pte_us = (vaddr >= USER_CODE_ADDR) ? PTE_US : 0;
	page_dir [page_dir_index] = make_PTE (page_alloc (), pte_us);
	FLUSH_TBL;
	memset (page_tbl_vaddr, 0, PAGE_SIZE);
    }

    /* if no page is allocated for 'vaddr', allocate it */
    pte_vaddr = (uint32_t *)((uint32_t)page_tbl_vaddr + (page_tbl_index << 2));
    if (!(*pte_vaddr & PTE_P)) {
	uint32_t pte_us = (vaddr >= USER_CODE_ADDR) ? PTE_US : 0;
	*pte_vaddr = make_PTE (page_alloc (), pte_us);
	FLUSH_TBL;
    }
}

void
free_pages   (int free_stacks)
{
    uint32_t *page_dir, *page_table;
    int i, j;

    /* xxx: hard coding like 511, 1024 is ugly */
    asm volatile ("movl %%cr3,    %0":"=r"(page_dir));
    for (i = 0; i < 1024; i++) {
	if (i == 0 || i == 1)
	    continue; /* page_dir [0] is IM area, page_dir [1] is PM area */
	if (i == 511 && !free_stacks)
	    continue; /* skip kernel stack */
	if (page_dir [i] & PTE_P) {
	    page_table = (uint32_t *)(PAGE_TBL_MAP_ADDR + (i << 12));
	    for (j = 0; j < 1024; j++) {
		if (page_table [j] & PTE_P) {
#if 1 /* for debug */
		    /* fill the page with some pattern for sanitary */
		    memset ((i << 22) | (j << 12), 0xF4, PAGE_SIZE); /* hlt */
#endif
		    page_free ((void *)(page_table [j] & 0xFFFFF000));
		}
	    }
	    page_free ((void *)(page_dir [i] & 0xFFFFF000));
	    page_dir [i] = 0;
	}
    }
    FLUSH_TBL;
}

uint32_t *
copy_pages (void)
{
    uint32_t *page_dir1,   *page_dir2 = im_page_alloc ();
    uint32_t *page_table1, *page_table2, *page1, *page2;
    int i, j;

    asm volatile ("movl %%cr3,    %0":"=r"(page_dir1));

    /* copy from page_dir1 to page_dir2 */
    memset (page_dir2, 0, PAGE_SIZE);
    page_dir2 [0] = page_dir1 [0];           /* share global page table */
    page_dir2 [1] = make_PTE (page_dir2, 0); /* map page tables */
    page_dir1 [2] = page_dir2 [1];           /* also map them to the current */
    FLUSH_TBL;

    /* 
       page_dir [0] = identical mapping area
       page_dir [1] = page table mapping area
       page_dir [2] = working area for copy_pages (map new page dir.)
       page_dir [3] = working area for copy_pages (map a new page)
     */
    for (i = 4; i < 1024; i++) {
	if (!(page_dir1 [i] & PTE_P))
	    continue;

	page_table1 = (uint32_t *)(PAGE_TBL_MAP_ADDR  + (i << 12));
	page_table2 = (uint32_t *)(PAGE_TBL_MAP_ADDR2 + (i << 12));

	page_dir2 [i] = make_PTE (page_alloc (), (i>=512 ? PTE_US : 0));

	for (j = 0; j < 1024; j++) {
	    if (!(page_table1 [j] & PTE_P)) {
		page_table2 [j] = 0;
		continue;
	    }
	    page_table2 [j] = make_PTE (page_alloc (), (i>=512 ? PTE_US : 0));
	    page_dir1 [3] = page_table2 [j]; /* map the page to the current */
	    FLUSH_TBL;
	    page1 = (void *)((i << 22) | (j << 12));
	    page2 = (uint32_t *)(PAGE_TBL_MAP_ADDR + (3 << 12));
	    memcpy (page2, page1, PAGE_SIZE);
	}
    }
    page_dir1 [2] = page_dir1 [3] = 0;
    FLUSH_TBL;
    return page_dir2;
}

void
free_kernel_stack (uint32_t cr3)
{
    int i, j;
    uint32_t *page_dir1, *page_dir2, *page_table2;

    asm volatile ("movl %%cr3,    %0":"=r"(page_dir1));
    page_dir2 = (uint32_t *) cr3;
#if 0
    printk ("page_dir1=%x, page_dir2=%x\n", page_dir1, page_dir2);
    for (i = 0; i < 1024; i++) {
	if ((page_dir1 [i] & PTE_P) || (page_dir2 [i] & PTE_P)) {
	    printk ("page_dir1[%d]=%x, page_dir2[%d]=%x\n",
		    i, page_dir1[i], i, page_dir2[i]);
	}
    }
#endif
    page_dir1 [2] = make_PTE (page_dir2, 0); /* map cr3's page tables */
    FLUSH_TBL;

    i = 511;
    assert (page_dir2 [i] & PTE_P);
    page_table2 = (uint32_t *)(PAGE_TBL_MAP_ADDR2 + (i << 12));

    for (j = 0; j < 1024; j++) {
	if (!(page_table2 [j] & PTE_P))
	    continue;
	page_free ((void *)(page_table2 [j] & 0xFFFFF000));
    }
    page_free ((void *)(page_dir2 [i] & 0xFFFFF000));
    im_page_free ((void *)((uint32_t) page_dir2 & 0xFFFFF000));
    page_dir1 [2] = page_dir1 [3] = 0;
    FLUSH_TBL;
}

void
dump_pages   (void)
{
    uint32_t *page_dir, *page_table;
    int i, j;

    asm volatile ("movl %%cr3,    %0":"=r"(page_dir));
    printk ("page_dir   =%x\n", page_dir);
    printk ("page_dir[0]=%x\n", page_dir [0]);
    printk ("page_dir[1]=%x\n", page_dir [1]);
    for (i = 2; i < 1024; i++) {
	if (page_dir [i] & PTE_P) {
	    printk ("page_dir[%d]=%x\n", i, page_dir [i]);
	    page_table = (uint32_t *)(PAGE_TBL_MAP_ADDR + (i << 12));
	    for (j = 0; j < 1024; j++) {
		if (page_table [j] & PTE_P)
		    printk ("   page_table[%d]=%x\n", j, page_table [j]);
	    }
	}
    }
    printk ("dump_pages done\n");
}

void
dump_mem_usage (void)
{
    struct CDLL *p;
    int i, j, size;

    /* kmalloc/kfree */
    size = 0;
    for (p = kmalloc_list.next; p != &kmalloc_list; p = p->next) {
	struct kmalloc_header *header = p->data;
	if (header->is_free)
	    size += header->size;
    }
    printk ("kmalloc: %x free, %x max\n", size,
	    KMALLOC_ADDR_END - KMALLOC_ADDR_START);

    /* im_page_alloc/im_page_free */
    size = 0;
    for (i = 0; i < IM_PAGE_NUM / 32; i++) {
	if (used_im_page_bitmap [i] != 0xFFFFFFFF) {
	    for (j = 0; j < 32; j++) {
		if (((1 << j) & used_im_page_bitmap [i]) == 0) {
		    size += PAGE_SIZE;
		}
	    }
	}
    }
    printk ("im_page_alloc: %x free, %x max\n", size, IM_MEM_SIZE);

    /* page_alloc/page_free */
    size = 0;
    for (i = 0; i < PAGE_NUM / 32; i++) {
	if (used_page_bitmap [i] != 0xFFFFFFFF) {
	    for (j = 0; j < 32; j++) {
		if (((1 << j) & used_page_bitmap [i]) == 0) {
		    size += PAGE_SIZE;
		}
	    }
	}
    }
    printk ("page_alloc: %x free, %x max\n", size, MEM_SIZE);
}
/* ---------------------------------------------------------------------- */
