/*
    udos: micro disk OS.
    Copyright (c) 2003, 2004   Katsuhiko Gondow (gondow@cs.titech.ac.jp)
    $Id: task.c,v 1.51 2004/07/14 05:43:19 gondow Exp gondow $
*/
#include "cpu.h"
#include "misc.h"
#include "task.h"
#include "syscall.h"
#include "paging.h" /* for create_aspace */
#include "boot32.h"
#include "fat12.h"
#include "list.h"
#include "fs.h"

#define SIG_DFL  ((void (*)(int))0)
#define SIG_IGN  ((void (*)(int))1)
/* ---------------------------------------------------------------------- */
struct process process_table [MAX_PROCESS];
jmp_buf syscall_jmp_buf;
/* ---------------------------------------------------------------------- */
static struct CDLL process_queue [6]; /* indexed by enum Process_Status */
static int ident_exit;
static int ident_sleep;
static void (*signal_handler_caller)(void) = NULL;
static int foreground_pid = -1;
/* ---------------------------------------------------------------------- */
static void create_task (struct TSS *tss, struct GDT *gdt, uint32_t *page_dir,
			 uint32_t eip, uint32_t esp, uint32_t ebp);
static struct process *move_queue (struct CDLL *cdll,
				   enum Process_Status status);
/* ---------------------------------------------------------------------- */
static void
create_task (struct TSS *tss, struct GDT *gdt, uint32_t *page_dir,
	     uint32_t eip, uint32_t esp, uint32_t ebp)
{
    /* initialize dynamic fields in TSS */
    tss->cs     = 0x08; /* kernel code selector */
    tss->eip    = eip;
    tss->ds = tss->ss = tss->es = tss->fs = tss->gs = 0x10; /* kernel data */
    tss->ldt    = 0x00; /* null selector */
    tss->esp    = esp;
    tss->ebp    = ebp;
    tss->eax    = 0;     /* return value for child process's sys_fork */
    tss->eflags = 0x202; /* bit[9](IF)=1, reserved as bit[1]=1 */

    /* initialize dynamic fields in TSS */
    tss->ss0    = 0x10; /* kernel data selector */
    tss->esp0   = KERNEL_STACK_ADDR;
    tss->cr3    = (uint32_t) page_dir;
    tss->ldt    = 0x00; /* null selector */
    tss->trace  = 0; /* no trap */
    tss->iobase = 0; /* empty I/O permission map */

    /* initialize GDT entry */
    {
    	uint32_t addr  = (uint32_t)tss;
	uint32_t limit = sizeof (struct TSS);
	gdt->word [0] = limit & 0xFFFF;
	gdt->word [1] = addr  & 0xFFFF;
	gdt->word [2] = 0x8900 | ((addr >> 16) & 0xFF);
	gdt->word [3] = ((addr  >> 16) & 0xFF00)
	              | ((0x0000)      & 0x00F0)
	              | ((limit >> 16) & 0x000F);
    }
}

/* must be called under klocked contexts */
static struct process *
move_queue (struct CDLL *cdll, enum Process_Status status)
{
    struct process *process = (struct process *)cdll->data;
    CDLL_remove (cdll);
    CDLL_insert_tail (&process_queue [status], cdll);
    process->status = status;
    return process;
}

/* ---------------------------------------------------------------------- */
void
dump_process_table (void)
{
    int i;
    struct CDLL *child;

    printk ("STATUS PID PPID IDENT TIMEOUT INTR SIGNAL    CHILDREN\n");
    for (i = 0; i < MAX_PROCESS; i++) {
	if (process_table [i].status == PROCESS_UNUSED)
	    continue;
	switch (process_table [i].status) {
	case PROCESS_RUNNING:  printk ("RUNNIN"); break;
	case PROCESS_READY:    printk ("READY "); break;
	case PROCESS_ZOMBIE:   printk ("ZOMBIE"); break;
	case PROCESS_BLOCKED:  printk ("BLOCK "); break;
	default:               assert (0);
	}
	printk ("   %d",  process_table [i].pid);
	printk ("    %d", process_table [i].ppid);
	if (process_table [i].sync_ident)
	    printk (" %x", process_table [i].sync_ident);
	else
	    printk (" -----");
	if (process_table [i].wakeup_time)
	    printk ("   %x", process_table [i].wakeup_time);
	else
	    printk (" -------");
	printk ("    %d", process_table [i].is_interruptible);
	printk (" %d:%x:%x:%x:%x",
		(check_signals (i) ? 1 : 0),
		process_table [i].signal[SIGSEGV].handler,
		process_table [i].signal[SIGINT ].handler,
		process_table [i].signal[SIGUSR1].handler,
		process_table [i].signal[SIGUSR2].handler
		);
	for (child = process_table [i].cdll_children.next;
	     child != &process_table [i].cdll_children; child = child->next ) {
	    struct process *process = (struct process *) child->data;
	    printk (" %d", process->pid);
	}
	printk ("\n");
    }
}

void
dump_stack (void)
{
    int i;
    uint32_t *ebp;
    asm volatile ("movl %%ebp, %0": "=r"(ebp));
    for (i = 0; i < 10; i++) {
	printk ("[%d] ebp=%x, eip=%x\n", i, ebp [0], ebp [1]);
	ebp = (uint32_t *) ebp [0];
	if (ebp == (uint32_t *) USER_STACK_ADDR)
	    break;
    }
}

void
tss_fault_handler (uint32_t error)
{
    printk ("tss fault: error-code = %x\n", error);
    dump_stack ();
    asm volatile ("cli; 1: hlt; jmp 1b");
}

void
gp_fault_handler (uint32_t error)
{
    printk ("general protection fault: error-code = %x\n", error);
    dump_stack ();
    asm volatile ("cli; 1: hlt; jmp 1b");
}

int
sys_exec (char *path, char *argv [])
{
    struct file file;
    char buf [512];
    int  ret, nbytes, i, argc;
    char *p;
    uint32_t *page;
    struct process *process = &process_table [sys_getpid ()];

    /* open the target file */
    ret = FAT12_open (&file, path, O_RDONLY);
    if (ret == -1) {
	printk ("sys_exec: No such filename = %s\n", path);
	return -1;
    }

    /* copy argv and its content to page */
    if (argv == NULL)
	goto free_pages;
    page = kmalloc (PAGE_SIZE);
    for (argc = 0; argv [argc] != NULL; argc++)
	;
    page [0] = argc;
    page [1] = (uint32_t)&((uint32_t *)USER_STACK_ADDR)[2];
    page [argc + 2] = (uint32_t) NULL;
    p = (char *)&page [argc + 2 + 1]; /* +1 for NULL */
    for (i = 0; i < argc; i++) {
	page [i + 2] = USER_STACK_ADDR + (p - (char *)page);
	assert (p - (char*)page < PAGE_SIZE);
	strcpy (p, argv [i]);
	p += strlen (argv [i]) + 1;
    }

 free_pages:
    free_pages (0); /* 0 means don't free kernel/user stacks[0] */

    /* setup user code area */
    p = (void *)USER_CODE_ADDR;
    while ((nbytes = FAT12_read (&file, buf, sizeof (buf))) > 0) {
	memcpy (p, buf, nbytes);
	p += nbytes;
    }
    FAT12_close (&file);

    if (argv == NULL)
	goto out;
    memcpy ((void *)USER_STACK_ADDR, page, PAGE_SIZE);
    kfree (page);

 out:
    /* set default signal handling */
    for (i = 0; i < MAX_SIGNAL; i++) {
	process->signal [i].handler = SIG_DFL;
    }

    /* create stack frame to imitate as if an interrupt hander is called. */
    /* iret requires DPL <= RPL, so the code selector must be 0x1B,
       not 0x18. So does the data selector. see p.363 in i386-2j.pdf */
    asm volatile ("pushl $0x23; pushl %0; pushl $0x202;"
		  "pushl $0x1B; pushl %1"::
		  "i"(USER_STACK_ADDR), "i"(USER_CODE_ADDR));
    asm volatile ("movw $0x23, %ax; movw %ax, %ds; movw %ax, %es;"
		  "movw %ax, %fs; movw %ax, %gs");
    asm volatile ("iret");

    return 0; /* dummy */
}

int
sys_fork (void)
{
    int i;
    uint32_t *cr3;
    uint32_t eip, esp, ebp, eflags;
    struct process *child, *self;

    /* first, get free process entry */
    if (CDLL_is_null (&process_queue [PROCESS_UNUSED]))
	panic ("sys_fork: process table full");

    self  = (struct process *) process_queue [PROCESS_RUNNING].next->data;
    klock2 (eflags);
    child = move_queue (process_queue [PROCESS_UNUSED].next, PROCESS_READY);
    kunlock2 (eflags);

    /* initialize the process structure */
    child->tss    = kmalloc (sizeof (struct TSS));
    child->ppid   = self->pid;
    klock ();
    child->wakeup_time = 0;
    child->sync_ident  = 0;
    child->current_dir = self->current_dir;
    kunlock ();

    child->cdll_self = CDLL_create (child);

    CDLL_wrap (&child->cdll_children, NULL);
    CDLL_insert_tail (&self->cdll_children, CDLL_create (child));

    for (i = 0; i < MAX_SIGNAL; i++) {
	child->signal [i].is_received = 0;
	child->signal [i].handler = self->signal [i].handler;
    }

    for (i = 0; i < MAX_PROCESS_FILE; i++) {
	child->file [i] = self->file [i];
	child->file [i]->ref_count++;
    }

    /* copy all pages */
    cr3 = copy_pages ();

    /* obtain eip, esp and ebp by examining the stack frame */
    asm volatile ("movl 4(%%ebp), %0":"=r"(eip)); /* return address */
    asm volatile ("leal 8(%%ebp), %0":"=r"(esp)); /* +8 to skip ebx/eip */
    asm volatile ("movl 0(%%ebp), %0":"=r"(ebp)); /* old ebp */
    
    /* initialize TSS (set eip&esp as if child just returns from sys_fork) */
    create_task (child->tss, &gdt [child->pid + 5], cr3, eip, esp, ebp);
    return child->pid;
}

void
sys_exit    (int status)
{
    int i;
    int pid = sys_getpid ();
    uint32_t eflags;

    /* kernel and init process MUST NOT call exit(); */
    assert (pid != 0 && pid != 1);

    /* invalidate foreground_pid */
    if (pid == foreground_pid)
	foreground_pid = -1;

    /* set the exit status */
    process_table [pid].exit_status = status;

    /* close all files */
    for (i = 0; i < MAX_PROCESS_FILE; i++)
	sys_close (i);

    /* get the child processes adopted to the init process (pid=1) */
    while (!CDLL_is_null (&process_table [pid].cdll_children)) {
	struct CDLL *cdll = process_table [pid].cdll_children.next;
	struct process *child = (struct process *) cdll->data;
	child->ppid = 1;
	CDLL_remove (cdll);
	CDLL_insert_tail (&process_table [1].cdll_children, cdll);
    }

    /* wakeup the parent process being blocked by wait() */
    klock2 (eflags);
    sync_notify (&ident_exit);
    kunlock2 (eflags);

    /* free almost most resources used in this process */
    free_pages (0);
    yield (PROCESS_ZOMBIE);
}

int
sys_wait    (int *status)
{
    uint32_t eflags;
    int pid;
    struct CDLL *p, *children = &process_table [sys_getpid ()].cdll_children;
    struct process *process;

 retry:
    if (CDLL_is_null (children)) {
	return -1;
    }

    for (p = children->next; p != children; p = p->next) {
	process = (struct process *)p->data;
	if (process->status == PROCESS_ZOMBIE) {
	    goto found_zombie;
	}
    }
    klock2 (eflags);
    sync_wait (&ident_exit, 0, INTERRUPTIBLE);
    process_signals (1);
    kunlock2 (eflags);
    goto retry;

 found_zombie:
    assert (process->status == PROCESS_ZOMBIE);
    pid = process->pid;

    /* free all resources in the zombie */
    free_kernel_stack (process_table [pid].tss->cr3);
    kfree (process_table [pid].tss);
    CDLL_remove (p);
    kfree (p);

    if (status != NULL)
	*status = process_table [pid].exit_status;
    klock2 (eflags);
    move_queue (process->cdll_self, PROCESS_UNUSED);
    kunlock2 (eflags);
    return pid;
}

int
sys_getpid (void)
{
    struct CDLL *cdll = process_queue [PROCESS_RUNNING].next;
    struct process *current = (struct process *)cdll->data;
    return current->pid;
}

int
sys_getppid (void)
{
    struct CDLL *cdll = process_queue [PROCESS_RUNNING].next;
    struct process *current = (struct process *)cdll->data;
    return current->ppid;
}

int
sys_msleep  (int ms)
{
    uint32_t eflags;
    klock2 (eflags);
    sync_wait (&ident_sleep, ms, INTERRUPTIBLE);
    process_signals (1);
    kunlock2 (eflags);
    return 0;
}

int
sys_chdir   (const char *path)
{
    int fd;
    struct process *process = &process_table [sys_getpid ()];

    fd = sys_open (path, O_RDONLY);
    if (fd == -1) {
	printk ("sys_chdir: No such file or directory: %s\n", path);
	return -1;
    }
    if (!is_directory (&process->file[fd]->dirent)) {
	printk ("sys_chdir: Not directory: %s\n", path);
	return -1;
    }
    process->current_dir = process->file [fd]->dirent.clus_num;
    sys_close (fd);
    return 0;
}

void (*
sys_signal (int sig_num, void (*handler)(int), void (*caller)(void)))(int)
{
    struct process *process = &process_table [sys_getpid ()];
    void (*old_handler)(int) = process->signal [sig_num].handler;
    process->signal [sig_num].handler = handler;
    signal_handler_caller = caller;
    return old_handler;
}

int
sys_kill    (int pid, int sig_num)
{
    uint32_t eflags;
    struct process *process = &process_table [pid];
    klock2 (eflags);
    if (pid == 0 || pid == 1
	|| process->status == PROCESS_UNUSED
	|| process->status == PROCESS_ZOMBIE) {
	kunlock2 (eflags);
	return -1;
    }
    assert (0 <= pid && pid < MAX_PROCESS);
    process->signal [sig_num].is_received = 1;
    kunlock2 (eflags);
    return 0;
}

int
sys_tcgetpid (int fd)
{
    /* currently, fd is ignored */
    return foreground_pid;
}

int
sys_tcsetpid (int fd, int pid)
{
    /* currently, fd is ignored */
    foreground_pid = pid;
    return 0;
}

int
system_call (int dummy1, int arg4, int arg5, int dummy2, int dummy3,
	     int arg1, int arg3, int arg2, int sys_num)
{
    uint32_t eflags;
    int ret;

    if (ksetjmp (syscall_jmp_buf) != 0)
	return -1;

    switch (sys_num) {
    case SYSCALL_FORK:  ret = sys_fork ();                            break;
    case SYSCALL_EXEC:  ret = sys_exec ((char *)arg1, (char **)arg2); break;
    case SYSCALL_EXIT:  sys_exit (arg1); ret = 0;                     break;
    case SYSCALL_WAIT:  ret = sys_wait ((int *)arg1);                 break;
    case SYSCALL_MSLEEP: ret = sys_msleep ((int)arg1);                break;
    case SYSCALL_OPEN:  ret = sys_open ((const char *)arg1, (int)arg2); break;
    case SYSCALL_CLOSE: ret = sys_close ((int)arg1);                  break;
    case SYSCALL_READ:  ret = sys_read ((int)arg1, (void *)arg2, (int)arg3);
	                                                              break;
    case SYSCALL_WRITE: ret = sys_write ((int)arg1, (void *)arg2, (int)arg3);
	                                                              break;
    case SYSCALL_SEEK:  ret = sys_seek ((int)arg1, (int)arg2, (int)arg3);
	                                                              break;
    case SYSCALL_READDIR: ret = sys_readdir ((int)arg1,
					    (struct FAT_dirent *)arg2); break;
    case SYSCALL_MKDIR:  ret = sys_mkdir ((const char *)arg1);        break;
    case SYSCALL_RMDIR:  ret = sys_rmdir ((const char *)arg1);        break;
    case SYSCALL_UNLINK: ret = sys_unlink ((const char *)arg1);       break;
    case SYSCALL_SYNC:   sys_sync (); ret = 0;                        break;
    case SYSCALL_CHDIR:  ret = sys_chdir ((const char *)arg1);        break;
    case SYSCALL_STAT:   ret = sys_stat ((const char *)arg1,
					(struct FAT_dirent *)arg2);   break;
    case SYSCALL_SIGNAL:
	ret = (int) sys_signal ((int)arg1, (void (*)(int))arg2,
			       (void (*)(void))arg3);                 break;
    case SYSCALL_KILL:   ret = sys_kill ((int)arg1, (int)arg2);       break;
    case SYSCALL_GETPID: ret = sys_getpid ();                         break;
    case SYSCALL_GETPPID: ret = sys_getppid ();                       break;
    case SYSCALL_TCGETPID: ret = sys_tcgetpid ((int)arg1);            break;
    case SYSCALL_TCSETPID: ret = sys_tcsetpid ((int)arg1, (int)arg2); break;


    case SYSCALL_PRINTK:
	printk ((const char*)arg1, arg2, arg3, arg4, arg5); ret = 0; break;
    case SYSCALL_PANIC:     _panic ((char *)arg1, (char *)arg2, (int)arg3);
	                                                    ret = 0; break;
    case SYSCALL_YIELD:              yield (PROCESS_READY); ret = 0; break;
    case SYSCALL_DUMP_MEM_USAGE:     dump_mem_usage ();     ret = 0; break;
    case SYSCALL_DUMP_PROCESS_TABLE: dump_process_table (); ret = 0; break;
    case SYSCALL_INFINITE_LOOP:      while (1);             ret = 0; break;
    case SYSCALL_SYNC_WAIT: 
	klock2 (eflags);
	ret = sync_wait ((void*)arg1, (int)arg2, UNINTERRUPTIBLE);
	kunlock2 (eflags);
	break;
    case SYSCALL_SYNC_NOTIFY:
	klock2 (eflags);
	sync_notify ((void *)arg1);
	kunlock2 (eflags);
	ret = 0;
	break;
    case SYSCALL_DUMP_STACK:
	dump_stack ();
	ret = 0;
	break;
    default:
	printk ("sys_num=%d\n", sys_num);
	panic ("sytem_call: undefined system call\n");
	break;
    }

    klock2 (eflags);
    if (yield_is_required) {
	yield_is_required = 0;
	yield (PROCESS_READY);
    }
    kunlock2 (eflags);
#if 0 /* xxx */
    process_signals (1);
#endif
    return ret;
}

void
task_init (void)
{
    int i;
    struct CDLL *cdll;

    /* initialize process queues */
    CDLL_wrap (&process_queue [PROCESS_UNUSED],  NULL);
    CDLL_wrap (&process_queue [PROCESS_RUNNING], NULL);
    CDLL_wrap (&process_queue [PROCESS_READY],   NULL);
    CDLL_wrap (&process_queue [PROCESS_ZOMBIE],  NULL);
    CDLL_wrap (&process_queue [PROCESS_BLOCKED], NULL);

    /* register the kernel task as the first process */
    cdll = CDLL_create (&process_table [0]);
    CDLL_insert_head (&process_queue [PROCESS_RUNNING], cdll);

    process_table [0].status = PROCESS_RUNNING;
    process_table [0].pid    = 0;
    process_table [0].ppid   = 0;
    process_table [0].tss    = &kernel_tss;
    process_table [0].wakeup_time = 0;
    process_table [0].sync_ident  = 0;
    process_table [0].current_dir = ROOT_DIR_CLUS_NUM;

    for (i = 0; i < MAX_SIGNAL; i++) {
	process_table [0].signal [i].is_received = 0;
	process_table [0].signal [i].handler     = SIG_DFL;
    }

    for (i = 0; i < MAX_PROCESS_FILE; i++) {
	process_table [0].file [i] = NULL;
    }

    process_table [0].cdll_self = cdll;
    CDLL_wrap (&process_table [0].cdll_children, NULL);

    /*register all the other processes as unused */
    for (i = 1; i < MAX_PROCESS; i++) {
	process_table [i].pid = PROCESS_UNUSED;
	process_table [i].pid = i;
	cdll = CDLL_create (&process_table [i]);
	CDLL_insert_tail (&process_queue [PROCESS_UNUSED], cdll);
    }
}

void
yield  (int status)
{
    struct process *running, *ready;
    uint32_t eflags, dest [2];

    klock2 (eflags);
    if (CDLL_is_null (&process_queue [PROCESS_READY])) {
	assert (sys_getpid () == 0);
	asm volatile ("sti;hlt");
	kunlock2 (eflags);
	return;
    }

    assert (!CDLL_is_null (&process_queue [PROCESS_RUNNING]));
    running = move_queue (process_queue [PROCESS_RUNNING].next, status);
    ready   = move_queue (process_queue [PROCESS_READY].next, PROCESS_RUNNING);

    dest [0] = 0x999; /* offset is not used for task switching */
    dest [1] = (ready->pid + 5) * 8; /* selector (assume little-endian) */

    asm volatile ("sti; ljmpl *%0"::"m"(*dest)); /* task switching */
    asm volatile ("clts"); /* clear CR0(TS) flag (see p.108 in i386-2j.pdf) */
    kunlock2 (eflags);
}

/* must be called under klocked contexts */
void
check_sleeping_processes (void)
{
    struct CDLL *p, *next;

    for (p = process_queue [PROCESS_BLOCKED].next;
	 p != &process_queue [PROCESS_BLOCKED]; p = next ) {
	struct process *process = (struct process *) p->data;
	next = p->next;
	if ((process->wakeup_time != 0 && PIT_ticks >= process->wakeup_time)
	    || (check_signals (process->pid) && process->is_interruptible)) {
	    /* wakeup sleeping process */
	    move_queue (p, PROCESS_READY);
	    process->wakeup_time = 0;
	    process->sync_ident  = 0;
	}
    }
}

/* must be called under klocked contexts */
int
sync_wait  (void *ident, int timeout_ms, int is_interruptible)
{
    struct CDLL *cdll = process_queue [PROCESS_RUNNING].next;
    struct process *process = (struct process *) cdll->data;
    process->sync_ident       = ident;
    process->is_interruptible = is_interruptible;
    if (timeout_ms != 0)
	process->wakeup_time = PIT_ticks + timeout_ms / PIT_MS;
    else
	process->wakeup_time = 0;
    yield (PROCESS_BLOCKED);
    /* return 0 for sync_notify, -1 for timeout. */
    return (process->sync_ident == ident ? 0 : -1);
}

/* must be called under klocked contexts */
void
sync_notify (void *ident)
{
    struct CDLL *p, *next;

    for (p = process_queue [PROCESS_BLOCKED].next;
	 p != &process_queue [PROCESS_BLOCKED]; p = next ) {
	struct process *process = (struct process *) p->data;
	next = p->next;
	if (process->sync_ident == ident) {
	    move_queue (p, PROCESS_READY);
	    process->wakeup_time = 0;
	}
    }
}

int
check_signals (int pid)
{
    int i;
    struct process *process = &process_table [pid];
    for (i = 0; i < MAX_SIGNAL; i++) {
	if (!process->signal [i].is_received)       continue;
	if (process->signal [i].handler == SIG_IGN) {
	    process->signal [i].is_received = 0;
	    continue;
	}
	return 1;
    }
    return 0;
}

void
process_signals (int do_klongjmp)
{
    int i;
    struct process *process = &process_table [sys_getpid ()];
    void (*handler)(int);
#if 0
    if (check_signals (sys_getpid ())) {
	printk ("process_signals: pid=%d\n", sys_getpid ());
	dump_process_table ();
    }
#endif

    for (i = 0; i < MAX_SIGNAL; i++) {
	if (!process->signal [i].is_received) continue;
	process->signal [i].is_received = 0;
	handler = process->signal [i].handler;
	if (handler == SIG_IGN) continue;
	if (handler == SIG_DFL) sys_exit (1);
#if 0
	printk ("pid=%d, handler=%x, sig=%d\n", sys_getpid (), handler, i);
#endif
	setup_signal_handler (handler, i);
	if (do_klongjmp)
	    klongjmp (syscall_jmp_buf, 1);
    }
}

/* modify user/kernel stack so that _signal_handler_caller is called
   when going back to user-mode. */
void
setup_signal_handler (void *handler, int signal_num)
{
    uint32_t *ebp, *esp;

    /* find the stack boundary between user/kernel stack */
    asm volatile ("movl %%ebp, %0": "=r"(ebp));
    while (1) {
	if ((uint32_t) ebp [0] >= USER_CODE_ADDR)
	    break;
	ebp = (void *)ebp [0];
    }
    esp = (uint32_t *) ebp [17];

    /* modify user-stack */
    esp [-1]  = ebp [16];  /* eflags */
    // esp [-1]  = 0x202;
    esp [-2]  = ebp [15];  /* cs  */
    esp [-3]  = ebp [14];  /* eip */
    esp [-4]  = ebp [10];  /* eax */
    esp [-5]  = ebp [ 9];  /* ecx */
    esp [-6]  = ebp [ 8];  /* edx */
    esp [-7]  = ebp [ 7];  /* ebx */
    esp [-8]  = (uint32_t) &esp [-2]; /* esp */
    esp [-9]  = ebp [ 5];  /* ebp */
    esp [-10] = ebp [ 4];  /* esi */
    esp [-11] = ebp [ 3];  /* edi */
    esp [-12] = signal_num;
    esp [-13] = (uint32_t) handler;

    /* modify kernel-stack */
    ebp [14] = (uint32_t) signal_handler_caller; /* eip */
    ebp [17] = (uint32_t) &esp [-13]; /* esp */
}
/* ---------------------------------------------------------------------- */
