/*
    udos: micro disk OS.
    Copyright (c) 2003, 2004   Katsuhiko Gondow (gondow@cs.titech.ac.jp)
    $Id: vga.c,v 1.27 2004/06/23 07:34:44 gondow Exp $
*/
/* ---------------------------------------------------------------------- */
#include "misc.h"
#include "vga.h"
#include "task.h"
#include "fd.h"    /* for motor_off_time */

/* C('A') produces ^A */
#define C(x)	   ((x) - 64)

/* VGA text screen info. */
enum {
    MAX_COLUMN = 80,
    MAX_ROW    = 25,
    TEXT_VRAM  = 0xB8000, /* color, text-mode */
    CODE       = 0,
    ATTR       = 1,
};

/* VGA text attributes */
enum { WHITE = 7, RED = 6, GREEN = 2, BLUE = 1, BLACK = 0, };

/* some scan codes */
enum {
    SCAN_CTRL   = 0x1D,
    SCAN_CAPS   = 0x3A,
    SCAN_LSHIFT = 0x2A,
    SCAN_RSHIFT = 0x36,
};
/* special keys. FULL means `keyboard buffer full' */
enum { FULL = -1, SHIFT = -2, CTRL = -3, ALT = -4, NONE = -5, CAPS = -6,
       ESC = 27, BS = 8, TAB = 9, RET = '\n', SPC = 32, };

/* keyboard */
enum { KBD_BUFSIZE = 4096 };
/* ---------------------------------------------------------------------- */
uint64_t PIT_ticks = 0;
uint16_t memory_size;
volatile int yield_is_required = 0;
/* ---------------------------------------------------------------------- */
static void vga_delete_char     (void);
static void kbd_put_char        (int c);
static void kbd_wait_IBF_0      (void);
static void kbd_wait_OBF_0      (void);
static void kbd_wait_OBF_1      (void);
static int  kbd_check_A20gate   (void);
/* ---------------------------------------------------------------------- */
static int curr_col = 0, curr_row = 0;
/*
  tvram is accessed like:
      tvram[10][20][CODE] = 'a';
      tvram[10][20][ATTR] = WHITE;
  This puts 'a' in white at the position where row=10 and column=20.
 */
static char (*tvram) [MAX_COLUMN][2] = (void *) TEXT_VRAM;
static int keymap [][128] = {
    /* not shifted */
    {FULL,ESC, '1', '2', '3', '4', '5', '6',      /*00-07*/
     '7', '8', '9', '0', '-', '=', BS,  TAB,      /*08-0F*/
     'q', 'w', 'e', 'r', 't', 'y', 'u', 'i',      /*10-17*/
     'o', 'p', '[', ']', RET, CTRL, 'a', 's',     /*18-1F*/
     'd', 'f', 'g', 'h', 'j', 'k', 'l', ';',      /*20-27*/
     '\'', '`', SHIFT, '\\', 'z', 'x', 'c', 'v',  /*28-2F*/
     'b', 'n', 'm', ',', '.', '/', SHIFT, NONE,   /*30-37*/
     ALT, SPC, CAPS},                  /* ^^^^ is PrtSc */
    /* shifted */
    {FULL,ESC, '!', '@', '#', '$', '%', '^',      /*00-07*/
     '&', '*', '(', ')', '_', '+', BS,  TAB,      /*08-0F*/
     'Q', 'W', 'E', 'R', 'T', 'Y', 'U', 'I',      /*10-17*/
     'O', 'P', '{', '}', RET, CTRL, 'A', 'S',     /*18-1F*/
     'D', 'F', 'G', 'H', 'J', 'K', 'L', ':',      /*20-27*/
     '\"', '~', SHIFT, '|', 'Z', 'X', 'C', 'V',   /*28-2F*/
     'B', 'N', 'M', '<', '>', '?', SHIFT, NONE,   /*30-37*/
     ALT, SPC, CAPS},
    /* ctrl pushed */
    {FULL,     NONE,   NONE,   NONE,   NONE,   NONE,   NONE,     30, /*00-07*/
     NONE,     NONE,   NONE,   NONE,     31,   NONE,   NONE,   NONE, /*08-0F*/
     C('Q'), C('W'), C('E'), C('R'), C('T'), C('Y'), C('U'), C('I'), /*10-17*/
     C('O'), C('P'),     27,     29,    RET,   CTRL, C('A'), C('S'), /*18-1F*/
     C('D'), C('F'), C('G'), C('H'), C('J'), C('K'), C('L'),   NONE, /*20-27*/
     NONE,     NONE,  SHIFT,     28, C('Z'), C('X'), C('C'), C('V'), /*28-2F*/
     C('B'), C('N'), C('M'),   NONE,   NONE,   NONE,  SHIFT,   NONE, /*30-37*/
        ALT,      0,  CAPS},
};

static char kbd_buf [KBD_BUFSIZE];    /* ring buffer */
static int  kbd_data_start = 0, kbd_data_end = 0; /* indices for kbd_buf */
static int  ident_kbd;
/* ---------------------------------------------------------------------- */
static void
vga_delete_char     (void)
{
    if (curr_col == 0 && curr_row == 0) {
	return;
    } else if (curr_col == 0 && curr_row > 0) {
	curr_col = MAX_COLUMN - 1;
	curr_row--;
    } else {
	curr_col--;
    }
    tvram [curr_row][curr_col][CODE] = ' ';
    tvram [curr_row][curr_col][ATTR] = WHITE;
}

/* must be called in the klock/kunlock context */
static void
kbd_put_char        (int c)
{
    if (c == C('M'))
	c = C('J');    /* CR -> LF conversion in input */

    if (c == C('C')) {
	int pid = sys_tcgetpid (0);
	if (pid != -1)
	    sys_kill (pid, SIGINT);
    } else if (c == C('V')) {
	dump_process_table ();
    } else if (c == C('H')) {
	if (kbd_data_start != kbd_data_end) {
	    kbd_data_end = (kbd_data_end + KBD_BUFSIZE - 1) % KBD_BUFSIZE;
	    vga_put_char (c);
	}
    } else if (c == C('I')) {
	int i;
	for (i = 0; i < 4; i++)
	    kbd_put_char (' ');
    } else if (c == C('L')) {
	vga_screen_clear ();
    } else if (c != C('J') && c < 32) {
	/* ignore other control characters */
    } else {
	kbd_buf [kbd_data_end] = c;
	kbd_data_end = (kbd_data_end + 1) % KBD_BUFSIZE;
	vga_put_char (c);
    }
    return;
}

static void
kbd_wait_IBF_0      (void)
{
    /* wait IBF (input buffer full, bit#1) becomes 0 */
    while ((inb (0x64) & 2) != 0)
	;    /* empty */
}

static void
kbd_wait_OBF_1      (void)
{
    /* wait OBF (output buffer full, bit#0) becomes 1 */
    while ((inb (0x64) & 1) == 0)
	;    /* empty */
}

static void
kbd_wait_OBF_0      (void)
{
    /* wait OBF (output buffer full, bit#0) becomes 0 */
    while ((inb (0x64) & 1) != 0)
	inb (0x60); /* discard data from KBC */
}

static int /* bool */
kbd_check_A20gate   (void)
{
    int tmp, result;
    int *p1 = (int *)0x8FE00, *p2 = (int *)0x18FE00;
    
    tmp = *p1;
    *p1 = 0xBEAF1111;
    *p2 = 0xBEAF2222;
    result = (*p1 != *p2); 
    *p1 = tmp;
    return result;
}
/* ---------------------------------------------------------------------- */
void
check_memory_size (void)
{
    asm volatile ("movw 0x7E00, %0":"=r"(memory_size));
    memory_size = memory_size / 1024 + 1;
    /* Note: over 64MB memory area is ignored */
}

void
delay (void)
{
    /* wait about 1us for slow I/O. 0x80 is a port of DMA page register.
       see also http://www.linux.or.jp/JF/JFdocs/IO-Port-Programming/
       high-resolution.html#AEN160 . */
    asm volatile ("outb %al, $0x80;");
}

void
outb (uint8_t value, uint16_t port)
{
    /* see also linux/include/asm-i386/io.h */
    asm volatile ("outb %0,%1"::"a"(value),"Nd"(port));
    delay ();
}

uint8_t
inb (uint16_t port)
{
    uint8_t value;
    /* see also linux/include/asm-i386/io.h */
    asm volatile ("inb  %1, %0":"=a"(value):"Nd"(port));
    delay ();
    return value;
}

void
outw (uint16_t value, uint16_t port)
{
    /* see also linux/include/asm-i386/io.h */
    asm volatile ("outw %0,%1"::"a"(value),"Nd"(port));
    delay ();
}

uint16_t
inw (uint16_t port)
{
    uint16_t value;
    /* see also linux/include/asm-i386/io.h */
    asm volatile ("inw  %1, %0":"=a"(value):"Nd"(port));
    delay ();
    return value;
}

void
kbd_enable_A20gate (void)
{
    int value;

    if (kbd_check_A20gate ()) {
	printk ("A20gate already enabled\n");
	return;	   /* A20 gate is already enabled */
    }

    /* read output port */
    kbd_wait_IBF_0 ();
    outb (0xD0, 0x64);  /* command read */
    kbd_wait_IBF_0 ();
    value = inb (0x60);

    /* write output port */
    kbd_wait_IBF_0 ();
    outb (0xD1, 0x64);  /* command write */
    kbd_wait_IBF_0 ();
#if 1
    outb (value | 2, 0x60);  /* enable A20 gate */
#else
    outb (value & ~2, 0x60);  /* disable A20 gate */
#endif
}

void
kbd_intr_handler (void)
{
    static int shift_ctrl = 0;
    int scancode;

    kbd_wait_OBF_1 ();
    scancode = inb (0x60);
    if (scancode == SCAN_LSHIFT || scancode == SCAN_RSHIFT) {
	shift_ctrl = 1;
    } else if (scancode == SCAN_LSHIFT + 0x80 
	       || scancode == SCAN_RSHIFT + 0x80) {
	shift_ctrl = 0;
    } else if (scancode == SCAN_CTRL || scancode == SCAN_CAPS) {
	shift_ctrl = 2;
    } else if (scancode == SCAN_CTRL + 0x80 || scancode == SCAN_CAPS + 0x80) {
	shift_ctrl = 0;
    } else if (scancode >= 0x80) {
	/* just skip break code */
    } else {
	int c = keymap[shift_ctrl][scancode];
	if (kbd_data_start != (kbd_data_end + 1) % KBD_BUFSIZE) {/* not full */
	    kbd_put_char (c);
	}
	sync_notify (&ident_kbd);
    }
    outb (0x20, 0x20); /* send EOI to PIC1 */
}

void
pit_intr_handler (void)
{
    uint32_t *ebp;

    outb (0x20, 0x20); /* send EOI to PIC1 */
    PIT_ticks++;

    if ((uint32_t) PIT_ticks % 2 != 0)
	return;

    /* check FDC motor */
    if (motor_off_time != 0 && PIT_ticks >= motor_off_time) {
	motor_off_time = 0;
	FDC_motor (0); /* motor off */
    }

    check_sleeping_processes ();
#if 1 /* xxx */
    process_signals (0);
#endif
    
    /* process switching */
    asm volatile ("movl 0(%%ebp), %0":"=r"(ebp));
    if (ebp [1] >= USER_CODE_ADDR) { /* user mode   */
	yield (PROCESS_READY);
    } else {                         /* kernel mode */
	yield_is_required = 1;       /* reserve yielding */
    }
}

void
setup_PIC8259A    (void)
{
    /* 0x11 = edge trigger, cascade, ICW4 required */
    outb (0x11, 0x20); /* send ICW1 to 8259A master */
    outb (0x11, 0xA0); /* send ICW1 to 8259A slave  */

    /* 0x20 = assign hardware interrupt # 0x20-27 for IRQ0-7. */
    outb (0x20, 0x21); /* send ICW2 to 8259A master */
    /* 0x28 = assign hardware interrupt # 0x28-2F for IRQ0-7. */
    outb (0x28, 0xA1); /* send ICW2 to 8259A slave */

    /* 0x04 = bitmap that shows where slaves are connected to the master. */
    outb (0x04, 0x21); /* send ICW3 to 8259A master */
    /* 0x02 = master's IR number that this slave is connected to */
    outb (0x02, 0xA1); /* send ICW3 to 8259A slave */

    /* 0x01 = 8086, normal fully nested, normal EOI (not AEOI) */
    outb (0x01, 0x21); /* send ICW4 to 8259A master */
    outb (0x01, 0xA1); /* send ICW4 to 8259A slave */

    /* 0xFF = mask off (inhibit) all interrupts for now */
    outb (0xFF, 0x21); /* send OCW1 to 8259A master */
    outb (0xFF, 0xA1); /* send OCW1 to 8259A slave */
}

void
setup_PIT8254     (uint16_t count)
{
    outb (0x36, 0x43);  /* counter 0, square-wave */
    outb (count & 0xFF00, 0x40); /* set lower 8 bits of initial count */
    outb ((count >> 8) & 0xFF00, 0x40); /* set upper 8 bits */
}

void
setup_KBD8042     (void)
{
#if 0
    /* state == 0x61 is obtained in Bochs-2.2.5 */
    uint8_t status;
    kbd_wait_IBF_0 ();
    outb (0x20, 0x64);  /* issume 'KBC read command' */
    kbd_wait_OBF_1 ();
    status = inb (0x60);
    printk ("KBC's status(1)=%x\n", status);
    asm ("1: hlt; jmp 1b");
#endif

    kbd_wait_IBF_0 ();
    kbd_wait_OBF_0 ();
    outb (0xFF, 0x60);	/* reset to keyboard */

    kbd_wait_IBF_0 ();
    kbd_wait_OBF_0 ();
    outb (0xF4, 0x60);  /* enable keyboard */

#if 0
    {
	uint8_t status;
	kbd_wait_IBF_0 ();
	kbd_wait_OBF_0 ();
	outb (0xF0, 0x60);  /* inquire scan code set */

	kbd_wait_OBF_1 ();
	status = inb (0x60);
	printk ("scan code =%x\n", status);

	kbd_wait_IBF_0 ();
	kbd_wait_OBF_0 ();
	outb (0x00, 0x60);  /* inquire scan code set */

	kbd_wait_OBF_1 ();
	status = inb (0x60);
	printk ("scan code =%x\n", status);

	kbd_wait_OBF_1 ();
	status = inb (0x60);
	printk ("scan code =%x\n", status);

	kbd_wait_OBF_1 ();
	status = inb (0x60);
	printk ("scan code =%x\n", status);
	asm ("1: hlt; jmp 1b");
    }
#endif
}

void
setup_RTC146818A  (void)
{
    int year, month, day, hour, minute, second;

    /* need to preserve 0x70's bit7, since it is NMI mask register setting */
#define RTC_send_byte(value, addr) \
    (outb ((inb (0x70) & 0x80) | addr, 0x70), outb (value, 0x71))
#define RTC_get_byte(addr)         \
    (outb ((inb (0x70) & 0x80) | addr, 0x70), inb  (0x71))
    
    RTC_send_byte (0x06, 0x0B); /* UTI=PIE=AIE=UIE=SQW=0, DF=0(BCD),
				   HF=1(24H), DSE=0(no daylight saving time) */
    /* Note:Bochs-2.0.2 does not support binary mode */

    year   = RTC_get_byte (0x09);
    month  = RTC_get_byte (0x08);
    day    = RTC_get_byte (0x07);
    hour   = RTC_get_byte (0x04);
    minute = RTC_get_byte (0x02);
    second = RTC_get_byte (0x00);

    /* this is local time */
    printk ("%x/%x/%x, %x:%x:%x\n", year, month, day, hour, minute, second);
#undef RTC_send_byte
#undef RTC_get_byte
}

void
vga_screen_clear  (void)
{
    int i, j;
    for (i = 0; i < MAX_ROW; i++) {
	for (j = 0; j < MAX_COLUMN; j++) {
	    tvram [i][j][CODE] = ' ';
	    tvram [i][j][ATTR] = WHITE;
	}
    }
    curr_col = curr_row = 0;
    vga_cursor_move ();
}

void vga_screen_scroll (void)
{
    int i, j;
    for (i = 0; i < MAX_ROW - 1; i++) {
	for (j = 0; j < MAX_COLUMN; j++) {
	    tvram [i][j][CODE] = tvram [i + 1][j][CODE]; 
	    tvram [i][j][ATTR] = tvram [i + 1][j][ATTR];
	}
    }
    for (j = 0; j < MAX_COLUMN; j++) {
	tvram [MAX_ROW - 1][j][CODE] = ' ';
	tvram [MAX_ROW - 1][j][ATTR] = WHITE;
    }
    curr_row--;
}

void
vga_cursor_move   (void)
{
    int location = MAX_COLUMN * curr_row + curr_col;
    
    outb (0x0E, 0x3D4);
    outb (location/256, 0x3D5);	   /* send higher byte */
    outb (0x0F, 0x3D4);
    outb (location%256, 0x3D5);    /* send lower byte */
}

void
vga_cursor_toggle (int on)
{
    if (on) {
	outb (0x0A, 0x3D4);
	outb (0,    0x3D5);
	outb (0x0B, 0x3D4);
	outb (15,   0x3D5);
    } else {
	outb (0x0A, 0x3D4);
	outb (32,   0x3D5);
    }
}

void
vga_put_char      (int c)
{
    switch (c) {
    case C('H'): vga_delete_char ();      break;
    case C('J'):
	curr_col = 0;
	curr_row++;
	if (curr_row == MAX_ROW) {
	    vga_screen_scroll ();
	}
	break;
    default:
	if (c < 32)
	    break;
	tvram [curr_row][curr_col][CODE] = c;
	tvram [curr_row][curr_col][ATTR] = WHITE;
	curr_col++;

	if (curr_col >= MAX_COLUMN) {
	    curr_col = 0;
	    curr_row++;
	}
	if (curr_row == MAX_ROW) {
	    vga_screen_scroll ();
	}
	break;
    }
    vga_cursor_move ();
}

void
vga_put_string    (char *s)
{
    while (*s != 0) {
	vga_put_char (*s);
	s++; 
    }
}

void
vga_put_signed_num    (int v, int radix)
{
    if (v < 0) {
	vga_put_char ('-');
	v *= -1;
    }
    vga_put_unsigned_num (v, radix);
}

void
vga_put_unsigned_num    (unsigned int v, int radix)
{
    enum {MAX = 16};
    int  i, h;
    int digit [MAX];

    if (v == 0) {
	vga_put_char ('0');
	return;
    }
    for (i = 0; v != 0; i++) {
	if (i >= MAX)
	    break;
	h = v % radix;
	digit [i] = ((h >= 10) ? ('A' + h - 10) : ('0' + h));
	v /= radix;
    }
    for (i--; i >= 0; i--) {
	vga_put_char (digit [i]);
    }
}

int
kbd_get_char (void)
{
    uint32_t eflags;
    int c, i;

    klock2 (eflags);
    while (1) {
#if 1   /* line buffering ('\n' or buffer full causes flushing */
	if (kbd_data_start == (kbd_data_end + 1) % KBD_BUFSIZE) /* full */
	    goto flush;
	for (i = kbd_data_start; i != kbd_data_end; i = (i+1) % KBD_BUFSIZE) {
	    assert ('\n' == C('J'));
	    if (kbd_buf [i] == '\n')
		goto flush;
	}
#else   /* no buffering */
	if (kbd_data_start != kbd_data_end)
	    break;
#endif
	sync_wait (&ident_kbd, 0, INTERRUPTIBLE);
	process_signals (1);
    }
 flush:
    c = kbd_buf [kbd_data_start];
    kbd_data_start = (kbd_data_start + 1) % KBD_BUFSIZE;
    kunlock2 (eflags);
    return c;
}

void
kbd_get_line (char *s)
{
    int c;
    while ((c = kbd_get_char ()) != '\n')
	*s++ = c;
    *s = '\0';
}

int
console_open      (struct file *fp, const char *path, int flag)
{
    /* do nothing */
    return 0;
}

int  console_close     (struct file *fp)
{
    /* do nothing */
    return 0;
}

int
console_read      (struct file *fp, void *buf, int count)
{
    int i;
    char *buf2 = buf;
    for (i = 0; i < count; i++) {
	buf2 [i] = kbd_get_char ();
    }
    return 0;
}

int
console_write     (struct file *fp, void *buf, int count)
{
    int i;
    char *buf2 = buf;
    for (i = 0; i < count; i++) {
	vga_put_char (buf2 [i]);
    }
    return 0;
}
/* ---------------------------------------------------------------------- */

